"""
    ulmo.ncdc.ghcn_daily.core
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    This module provides direct access to `National Climatic Data Center`_
    `Global Historical Climate Network - Daily`_ dataset.


    .. _National Climatic Data Center: http://www.ncdc.noaa.gov
    .. _Global Historical Climate Network - Daily: http://www.ncdc.noaa.gov/oa/climate/ghcn-daily/

"""
from builtins import str
from builtins import range
from past.builtins import basestring
import itertools
import os

import numpy as np
import pandas

from ulmo import util


GHCN_DAILY_DIR = os.path.join(util.get_ulmo_dir(), 'ncdc/ghcn_daily')


def get_data(station_id, elements=None, update=True, as_dataframe=False):
    """Retrieves data for a given station.


    Parameters
    ----------
    station_id : str
        Station ID to retrieve data for.
    elements : ``None``, str, or list of str
        If specified, limits the query to given element code(s).
    update : bool
        If ``True`` (default),  new data files will be downloaded if they are
        newer than any previously cached files. If ``False``, then previously
        downloaded files will be used and new files will only be downloaded if
        there is not a previously downloaded file for a given station.
    as_dataframe : bool
        If ``False`` (default), a dict with element codes mapped to value dicts
        is returned. If ``True``, a dict with element codes mapped to equivalent
        pandas.DataFrame objects will be returned. The pandas dataframe is used
        internally, so setting this to ``True`` is a little bit faster as it
        skips a serialization step.


    Returns
    -------
    site_dict : dict
        A dict with element codes as keys, mapped to collections of values. See
        the ``as_dataframe`` parameter for more.
    """
    if isinstance(elements, basestring):
        elements = [elements]

    start_columns = [
        ('year', 11, 15, int),
        ('month', 15, 17, int),
        ('element', 17, 21, str),
    ]
    value_columns = [
        ('value', 0, 5, float),
        ('mflag', 5, 6, str),
        ('qflag', 6, 7, str),
        ('sflag', 7, 8, str),
    ]
    columns = list(itertools.chain(start_columns, *[
        [(name + str(n), start + 13 + (8 * n), end + 13 + (8 * n), converter)
         for name, start, end, converter in value_columns]
        for n in range(1, 32)
    ]))

    station_file_path = _get_ghcn_file(
        station_id + '.dly', check_modified=update)
    station_data = util.parse_fwf(station_file_path, columns, na_values=[-9999])

    dataframes = {}

    for element_name, element_df in station_data.groupby('element'):
        if not elements is None and element_name not in elements:
            continue

        element_df['month_period'] = element_df.apply(
                lambda x: pandas.Period('%s-%s' % (x['year'], x['month'])),
                axis=1)
        element_df = element_df.set_index('month_period')
        monthly_index = element_df.index

        # here we're just using pandas' builtin resample logic to construct a daily
        # index for the timespan
        # 2018/11/27 johanneshorak: hotfix to get ncdc ghcn_daily working again
        # new resample syntax requires resample method to generate resampled index.
        daily_index = element_df.resample('D').sum().index.copy()

        # XXX: hackish; pandas support for this sort of thing will probably be
        # added soon
        month_starts = (monthly_index - 1).asfreq('D') + 1
        dataframe = pandas.DataFrame(
                columns=['value', 'mflag', 'qflag', 'sflag'], index=daily_index)

        for day_of_month in range(1, 32):
            dates = [date for date in (month_starts + day_of_month - 1)
                    if date.day == day_of_month]
            if not len(dates):
                continue
            months = pandas.PeriodIndex([pandas.Period(date, 'M') for date in dates])
            for column_name in dataframe.columns:
                col = column_name + str(day_of_month)
                dataframe[column_name][dates] = element_df[col][months].values

        dataframes[element_name] = dataframe

    if as_dataframe:
        return dataframes
    else:
        return dict([
            (key, util.dict_from_dataframe(dataframe))
            for key, dataframe in dataframes.items()
        ])


def get_stations(country=None, state=None, elements=None, start_year=None,
        end_year=None, update=True, as_dataframe=False):
    """Retrieves station information, optionally limited to specific parameters.


    Parameters
    ----------
    country : str
        The country code to use to limit station results. If set to ``None``
        (default), then stations from all countries are returned.
    state : str
        The state code to use to limit station results. If set to ``None``
        (default), then stations from all states are returned.
    elements : ``None``, str, or list of str
        If specified, station results will be limited to the given element codes
        and only stations that have data for any these elements will be
        returned.
    start_year : int
        If specified, station results will be limited to contain only stations
        that have data after this year. Can be combined with the ``end_year``
        argument to get stations with data within a range of years.
    end_year : int
        If specified, station results will be limited to contain only stations
        that have data before this year. Can be combined with the ``start_year``
        argument to get stations with data within a range of years.
    update : bool
        If ``True`` (default),  new data files will be downloaded if they are
        newer than any previously cached files. If ``False``, then previously
        downloaded files will be used and new files will only be downloaded if
        there is not a previously downloaded file for a given station.
    as_dataframe : bool
        If ``False`` (default), a dict with station IDs keyed to station dicts
        is returned. If ``True``, a single pandas.DataFrame object will be
        returned. The pandas dataframe is used internally, so setting this to
        ``True`` is a little bit faster as it skips a serialization step.


    Returns
    -------
    stations_dict : dict or pandas.DataFrame
        A dict or pandas.DataFrame representing station information for stations
        matching the arguments. See the ``as_dataframe`` parameter for more.
    """

    columns = [
        ('country', 0, 2, None),
        ('network', 2, 3, None),
        ('network_id', 3, 11, None),
        ('latitude', 12, 20, None),
        ('longitude', 21, 30, None),
        ('elevation', 31, 37, None),
        ('state', 38, 40, None),
        ('name', 41, 71, None),
        ('gsn_flag', 72, 75, None),
        ('hcn_flag', 76, 79, None),
        ('wm_oid', 80, 85, None),
    ]

    stations_file = _get_ghcn_file('ghcnd-stations.txt', check_modified=update)
    stations = util.parse_fwf(stations_file, columns)

    if not country is None:
        stations = stations[stations['country'] == country]
    if not state is None:
        stations = stations[stations['state'] == state]

    # set station id and index by it
    stations['id'] = stations[['country', 'network', 'network_id']].T.apply(''.join)

    if not elements is None or not start_year is None or not end_year is None:
        inventory = _get_inventory(update=update)
        if not elements is None:
            if isinstance(elements, basestring):
                elements = [elements]

            mask = np.zeros(len(inventory), dtype=bool)
            for element in elements:
                mask += inventory['element'] == element
            inventory = inventory[mask]
        if not start_year is None:
            inventory = inventory[inventory['last_year'] >= start_year]
        if not end_year is None:
            inventory = inventory[inventory['first_year'] <= end_year]

        uniques = inventory['id'].unique()
        ids = pandas.DataFrame(uniques, index=uniques, columns=['id'])
        stations = pandas.merge(stations, ids).set_index('id', drop=False)

    stations = stations.set_index('id', drop=False)
    # wm_oid gets convertidsed as a float, so cast it to str manually
    # pandas versions prior to 0.13.0 could use numpy's fix-width string type
    # to do this but that stopped working in pandas 0.13.0 - fortunately a
    # regex-based helper method was added then, too
    if pandas.__version__ < '0.13.0':
        stations['wm_oid'] = stations['wm_oid'].astype('|U5')
        stations['wm_oid'][stations['wm_oid'] == 'nan'] = np.nan
    else:
        stations['wm_oid'] = stations['wm_oid'].astype('|U5').map(lambda x: x[:-2])
        is_nan = stations['wm_oid'] == 'n'
        is_empty = stations['wm_oid'] == ''
        is_invalid = is_nan | is_empty
        stations.loc[is_invalid, 'wm_oid'] = np.nan

    if as_dataframe:
        return stations
    else:
        return util.dict_from_dataframe(stations)


def _get_ghcn_file(filename, check_modified=True):
    base_url = 'http://www1.ncdc.noaa.gov/pub/data/ghcn/daily/'
    if 'ghcnd-' in filename:
        url = base_url + filename
    else:
        url = base_url + 'all/' + filename

    path = os.path.join(GHCN_DAILY_DIR, url.split('/')[-1])
    util.download_if_new(url, path, check_modified=check_modified)
    return path


def _get_inventory(update=True):
    columns = [
        ('id', 0, 11, None),
        ('latitude', 12, 20, None),
        ('longitude', 21, 30, None),
        ('element', 31, 35, None),
        ('first_year', 36, 40, None),
        ('last_year', 41, 45, None),
    ]

    inventory_file = _get_ghcn_file('ghcnd-inventory.txt',
            check_modified=update)
    return util.parse_fwf(inventory_file, columns)



"""
    `National Elevation Dataset (NED)`_ services (Raster)

    .. _National Elevation Dataset (NED): http://ned.usgs.gov
"""
from .core import get_available_layers, get_raster, get_raster_availability




"""
    ulmo.usace.rivergages.core
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module provides access to data provided by the `United States Army
    Corps of Engineers`_ `Rivergages`_ web site.

    .. _United States Army Corps of Engineers: http://www.usace.army.mil/
    .. _Rivergages: http://rivergages.mvr.usace.army.mil/WaterControl/new/layout.cfm
"""
import datetime
import os.path

import requests
from bs4 import BeautifulSoup

from ulmo import util

USACE_RIVERGAGES_DIR = os.path.join(util.get_ulmo_dir(), 'usace/rivergages/')
URL = 'http://rivergages.mvr.usace.army.mil/WaterControl/datamining2.cfm'
DEFAULT_START_DATE = datetime.date(1800, 1, 1)


def get_stations():
    path = os.path.join(USACE_RIVERGAGES_DIR, 'datamining_field_list.cfm')

    with util.open_file_for_url(URL, path, use_bytes=True) as f:
        soup = BeautifulSoup(f)
        options = soup.find('select', id='fld_station').find_all('option')
        stations = _parse_options(options)

    return stations


def get_station_data(station_code, parameter, start=None, end=None,
        min_value=None, max_value=None):

    if min_value is None:
        min_value = -9000000
    if max_value is None:
        max_value = 9000000
    if start is None:
        start_date = DEFAULT_START_DATE
    else:
        start_date = util.convert_date(start)
    if end is None:
        end_date = datetime.date.today()
    else:
        end_date = util.convert_date(end)

    start_date_str = _format_date(start_date)
    end_date_str = _format_date(end_date)

    form_data = {
        'fld_station': station_code,
        'fld_parameter': parameter,
        'fld_from': min_value,
        'fld_to': max_value,
        'fld_fromdate': start_date_str,
        'fld_todate': end_date_str,
        'hdn_excel': '',
    }

    req = requests.post(URL, params=dict(sid=station_code), data=form_data)
    soup = BeautifulSoup(req.content)
    data_table = soup.find('table').find_all('table')[-1]

    return dict([
        _parse_value(value_tr)
        for value_tr in data_table.find_all('tr')[2:]
    ])


def get_station_parameters(station_code):
    req = requests.get(URL, params=dict(sid=station_code))
    soup = BeautifulSoup(req.content)

    options = soup.find('select', id='fld_parameter').find_all()
    parameters = _parse_options(options)
    return parameters


def _format_date(date):
    return '%s/%s/%s' % (date.month, date.day, date.year)


def _parse_options(options):
    return dict([
        (option.attrs.get('value'), option.text.strip())
        for option in options
        if option.attrs.get('value') != ''
    ])


def _parse_value(value_tr):
    date_td, value_td = value_tr.find_all('td')

    return (util.convert_date(date_td.text),
        float(value_td.text))




import datetime

from ulmo import util


def test_convert_date_from_string():
    compare_dates = [
        ('2011-12-31', datetime.date(2011, 12, 31)),
        ('12/31/2011', datetime.date(2011, 12, 31)),
        ('2012-02-29', datetime.date(2012, 2, 29)),
        ('2012-2-29', datetime.date(2012, 2, 29)),
        ('2/29/2012', datetime.date(2012, 2, 29)),
        ('02/29/2012', datetime.date(2012, 2, 29)),
        ('2013-01-01', datetime.date(2013, 1, 1)),
    ]

    for test_str, test_date in compare_dates:
        date = util.convert_date(test_str)
        assert date == test_date


def test_convert_date_from_date():
    compare_dates = [
        datetime.date(2011, 12, 31),
        datetime.date(2012, 2, 29),
        datetime.date(2013, 1, 1),
    ]

    for test_date in compare_dates:
        date = util.convert_date(test_date)
        assert date == test_date


def test_convert_date_from_datetime():
    compare_dates = [
        (datetime.datetime(2011, 12, 31, 20),
            datetime.date(2011, 12, 31)),
        (datetime.datetime(2011, 12, 31, 0, 0, 0),
            datetime.date(2011, 12, 31)),
        (datetime.datetime(2011, 12, 31, 23, 59, 59),
            datetime.date(2011, 12, 31)),
    ]

    for test_datetime, test_date in compare_dates:
        date = util.convert_date(test_datetime)
        assert date == test_date


def test_convert_datetime_from_string():
    compare_datetimes = [
        ('2011-12-31', datetime.datetime(2011, 12, 31)),
        ('2011-12-31 4:28', datetime.datetime(2011, 12, 31, 4, 28)),
        ('2011-12-31 4:28:15', datetime.datetime(2011, 12, 31, 4, 28, 15)),
        ('12/31/2011', datetime.datetime(2011, 12, 31)),
        ('12/31/2011 1:30:29', datetime.datetime(2011, 12, 31, 1, 30, 29)),
        ('12/31/2011 01:30:29', datetime.datetime(2011, 12, 31, 1, 30, 29)),
        ('12/31/2011 01:30', datetime.datetime(2011, 12, 31, 1, 30, 0)),
        ('2012-02-29', datetime.datetime(2012, 2, 29)),
        ('2012-2-29', datetime.datetime(2012, 2, 29)),
        ('2/29/2012', datetime.datetime(2012, 2, 29)),
        ('02/29/2012', datetime.datetime(2012, 2, 29)),
        ('2013-01-01', datetime.datetime(2013, 1, 1)),
    ]

    for test_str, test_datetime in compare_datetimes:
        converted = util.convert_datetime(test_str)
        assert converted == test_datetime


def test_convert_datetime_from_datetime():
    compare_datetimes = [
        datetime.datetime(2011, 12, 31),
        datetime.datetime(2011, 12, 31, 3, 30),
        datetime.datetime(2011, 12, 31, 1, 30, 19),
        datetime.datetime(2011, 12, 31, 0, 0, 1),
        datetime.datetime(2012, 2, 29),
        datetime.datetime(2013, 1, 1),
    ]

    for test_datetime in compare_datetimes:
        converted = util.convert_datetime(test_datetime)
        assert converted == test_datetime


def test_convert_datetime_from_date():
    compare_datetimes = [
        (datetime.date(2011, 12, 31),
            datetime.datetime(2011, 12, 31, 0, 0, 0)),
    ]

    for test_date, test_datetime in compare_datetimes:
        converted = util.convert_datetime(test_date)
        assert converted == test_datetime



"""
    ulmo.twc.kbdi.core
    ~~~~~~~~~~~~~~~~~~~~~

    This module provides direct access to `Texas Weather Connection`_ `Daily
    Keetch-Byram Drought Index (KBDI)`_ dataset.

    .. _Texas Weather Connection: http://twc.tamu.edu/
    .. _Daily Keetch-Byram Drought Index (KBDI): http://twc.tamu.edu/drought/kbdi
"""

import datetime
import os

import numpy as np
import pandas

from ulmo import util

CSV_SWITCHOVER = pandas.Timestamp('2016-10-01')

def get_data(county=None, start=None, end=None, as_dataframe=False, data_dir=None):
    """Retreives data.

    Parameters
    ----------
    county : ``None`` or str
        If specified, results will be limited to the county corresponding to the
        given 5-character Texas county fips code i.e. 48???.
    end : ``None`` or date (see :ref:`dates-and-times`)
        Results will be limited to data on or before this date. Default is the
        current date.
    start : ``None`` or date (see :ref:`dates-and-times`)
        Results will be limited to data on or after this date. Default is the
        start of the calendar year for the end date.
    as_dataframe: bool
        If ``False`` (default), a dict with a nested set of dicts will be
        returned with data indexed by 5-character Texas county FIPS code. If ``True``
        then a pandas.DataFrame object will be returned.  The pandas dataframe
        is used internally, so setting this to ``True`` is a little bit faster
        as it skips a serialization step.
    data_dir : ``None`` or directory path
        Directory for holding downloaded data files. If no path is provided
        (default), then a user-specific directory for holding application data
        will be used (the directory will depend on the platform/operating
        system).


    Returns
    -------
    data : dict or pandas.Dataframe
        A dict or pandas.DataFrame representing the data. See the
        ``as_dataframe`` parameter for more.
    """
    if end is None:
        end_date = datetime.date.today()
    else:
        end_date = util.convert_date(end)
    if start is None:
        start_date = datetime.date(end_date.year, 1, 1)
    else:
        start_date = util.convert_date(start)
    if data_dir is None:
        data_dir = os.path.join(util.get_ulmo_dir(), 'twc/kbdi')

    df = pandas.concat([
        _date_dataframe(date, data_dir)
        for date in pandas.period_range(start_date, end_date, freq='D')
    ], ignore_index=True)
    fips_df = _fips_dataframe()
    df = pandas.merge(df, fips_df, left_on='county', right_on='name')
    del df['name']

    if county:
        df = df[df['fips'] == county]

    if as_dataframe:
        return df
    else:
        return _as_data_dict(df)


def _as_data_dict(df):
    df['date'] = df['date'].map(str)
    county_dict = {}
    for county in df['fips'].unique():
        county_df = df[df['fips'] == county]
        county_data = county_df.T.drop(['fips'])
        values = [v.to_dict() for k, v in county_data.iteritems()]
        county_dict[county] = values

    return county_dict


def _date_dataframe(date, data_dir):

    if date.to_timestamp() < CSV_SWITCHOVER:
        url = _get_text_url(date)
        with _open_data_file(url, data_dir) as data_file:
            date_df = _parse_text_file(data_file)
    else:
        url = _get_csv_url(date)
        with _open_data_file(url, data_dir) as data_file:
            date_df = _parse_csv_file(data_file)

    date_df['date'] = pandas.Period(date, freq='D')

    return date_df


def _fips_dataframe():
    # fips codes from http://www.census.gov/geo/www/ansi/national.txt
    # with names adjusted to match twc kbdi: DEWITT --> DE WITT
    codes = (
        ('ANDERSON', 48001),
        ('ANDREWS', 48003),
        ('ANGELINA', 48005),
        ('ARANSAS', 48007),
        ('ARCHER', 48009),
        ('ARMSTRONG', 48011),
        ('ATASCOSA', 48013),
        ('AUSTIN', 48015),
        ('BAILEY', 48017),
        ('BANDERA', 48019),
        ('BASTROP', 48021),
        ('BAYLOR', 48023),
        ('BEE', 48025),
        ('BELL', 48027),
        ('BEXAR', 48029),
        ('BLANCO', 48031),
        ('BORDEN', 48033),
        ('BOSQUE', 48035),
        ('BOWIE', 48037),
        ('BRAZORIA', 48039),
        ('BRAZOS', 48041),
        ('BREWSTER', 48043),
        ('BRISCOE', 48045),
        ('BROOKS', 48047),
        ('BROWN', 48049),
        ('BURLESON', 48051),
        ('BURNET', 48053),
        ('CALDWELL', 48055),
        ('CALHOUN', 48057),
        ('CALLAHAN', 48059),
        ('CAMERON', 48061),
        ('CAMP', 48063),
        ('CARSON', 48065),
        ('CASS', 48067),
        ('CASTRO', 48069),
        ('CHAMBERS', 48071),
        ('CHEROKEE', 48073),
        ('CHILDRESS', 48075),
        ('CLAY', 48077),
        ('COCHRAN', 48079),
        ('COKE', 48081),
        ('COLEMAN', 48083),
        ('COLLIN', 48085),
        ('COLLINGSWORTH', 48087),
        ('COLORADO', 48089),
        ('COMAL', 48091),
        ('COMANCHE', 48093),
        ('CONCHO', 48095),
        ('COOKE', 48097),
        ('CORYELL', 48099),
        ('COTTLE', 48101),
        ('CRANE', 48103),
        ('CROCKETT', 48105),
        ('CROSBY', 48107),
        ('CULBERSON', 48109),
        ('DALLAM', 48111),
        ('DALLAS', 48113),
        ('DAWSON', 48115),
        ('DE WITT', 48123),
        ('DEAF SMITH', 48117),
        ('DELTA', 48119),
        ('DENTON', 48121),
        ('DEWITT', 48123),
        ('DICKENS', 48125),
        ('DIMMIT', 48127),
        ('DONLEY', 48129),
        ('DUVAL', 48131),
        ('EASTLAND', 48133),
        ('ECTOR', 48135),
        ('EDWARDS', 48137),
        ('EL PASO', 48141),
        ('ELLIS', 48139),
        ('ERATH', 48143),
        ('FALLS', 48145),
        ('FANNIN', 48147),
        ('FAYETTE', 48149),
        ('FISHER', 48151),
        ('FLOYD', 48153),
        ('FOARD', 48155),
        ('FORT BEND', 48157),
        ('FRANKLIN', 48159),
        ('FREESTONE', 48161),
        ('FRIO', 48163),
        ('GAINES', 48165),
        ('GALVESTON', 48167),
        ('GARZA', 48169),
        ('GILLESPIE', 48171),
        ('GLASSCOCK', 48173),
        ('GOLIAD', 48175),
        ('GONZALES', 48177),
        ('GRAY', 48179),
        ('GRAYSON', 48181),
        ('GREGG', 48183),
        ('GRIMES', 48185),
        ('GUADALUPE', 48187),
        ('HALE', 48189),
        ('HALL', 48191),
        ('HAMILTON', 48193),
        ('HANSFORD', 48195),
        ('HARDEMAN', 48197),
        ('HARDIN', 48199),
        ('HARRIS', 48201),
        ('HARRISON', 48203),
        ('HARTLEY', 48205),
        ('HASKELL', 48207),
        ('HAYS', 48209),
        ('HEMPHILL', 48211),
        ('HENDERSON', 48213),
        ('HIDALGO', 48215),
        ('HILL', 48217),
        ('HOCKLEY', 48219),
        ('HOOD', 48221),
        ('HOPKINS', 48223),
        ('HOUSTON', 48225),
        ('HOWARD', 48227),
        ('HUDSPETH', 48229),
        ('HUNT', 48231),
        ('HUTCHINSON', 48233),
        ('IRION', 48235),
        ('JACK', 48237),
        ('JACKSON', 48239),
        ('JASPER', 48241),
        ('JEFF DAVIS', 48243),
        ('JEFFERSON', 48245),
        ('JIM HOGG', 48247),
        ('JIM WELLS', 48249),
        ('JOHNSON', 48251),
        ('JONES', 48253),
        ('KARNES', 48255),
        ('KAUFMAN', 48257),
        ('KENDALL', 48259),
        ('KENEDY', 48261),
        ('KENT', 48263),
        ('KERR', 48265),
        ('KIMBLE', 48267),
        ('KING', 48269),
        ('KINNEY', 48271),
        ('KLEBERG', 48273),
        ('KNOX', 48275),
        ('LA SALLE', 48283),
        ('LAMAR', 48277),
        ('LAMB', 48279),
        ('LAMPASAS', 48281),
        ('LAVACA', 48285),
        ('LEE', 48287),
        ('LEON', 48289),
        ('LIBERTY', 48291),
        ('LIMESTONE', 48293),
        ('LIPSCOMB', 48295),
        ('LIVE OAK', 48297),
        ('LLANO', 48299),
        ('LOVING', 48301),
        ('LUBBOCK', 48303),
        ('LYNN', 48305),
        ('MADISON', 48313),
        ('MARION', 48315),
        ('MARTIN', 48317),
        ('MASON', 48319),
        ('MATAGORDA', 48321),
        ('MAVERICK', 48323),
        ('MCCULLOCH', 48307),
        ('MCLENNAN', 48309),
        ('MCMULLEN', 48311),
        ('MEDINA', 48325),
        ('MENARD', 48327),
        ('MIDLAND', 48329),
        ('MILAM', 48331),
        ('MILLS', 48333),
        ('MITCHELL', 48335),
        ('MONTAGUE', 48337),
        ('MONTGOMERY', 48339),
        ('MOORE', 48341),
        ('MORRIS', 48343),
        ('MOTLEY', 48345),
        ('NACOGDOCHES', 48347),
        ('NAVARRO', 48349),
        ('NEWTON', 48351),
        ('NOLAN', 48353),
        ('NUECES', 48355),
        ('OCHILTREE', 48357),
        ('OLDHAM', 48359),
        ('ORANGE', 48361),
        ('PALO PINTO', 48363),
        ('PANOLA', 48365),
        ('PARKER', 48367),
        ('PARMER', 48369),
        ('PECOS', 48371),
        ('POLK', 48373),
        ('POTTER', 48375),
        ('PRESIDIO', 48377),
        ('RAINS', 48379),
        ('RANDALL', 48381),
        ('REAGAN', 48383),
        ('REAL', 48385),
        ('RED RIVER', 48387),
        ('REEVES', 48389),
        ('REFUGIO', 48391),
        ('ROBERTS', 48393),
        ('ROBERTSON', 48395),
        ('ROCKWALL', 48397),
        ('RUNNELS', 48399),
        ('RUSK', 48401),
        ('SABINE', 48403),
        ('SAN AUGUSTINE', 48405),
        ('SAN JACINTO', 48407),
        ('SAN PATRICIO', 48409),
        ('SAN SABA', 48411),
        ('SCHLEICHER', 48413),
        ('SCURRY', 48415),
        ('SHACKELFORD', 48417),
        ('SHELBY', 48419),
        ('SHERMAN', 48421),
        ('SMITH', 48423),
        ('SOMERVELL', 48425),
        ('STARR', 48427),
        ('STEPHENS', 48429),
        ('STERLING', 48431),
        ('STONEWALL', 48433),
        ('SUTTON', 48435),
        ('SWISHER', 48437),
        ('TARRANT', 48439),
        ('TAYLOR', 48441),
        ('TERRELL', 48443),
        ('TERRY', 48445),
        ('THROCKMORTON', 48447),
        ('TITUS', 48449),
        ('TOM GREEN', 48451),
        ('TRAVIS', 48453),
        ('TRINITY', 48455),
        ('TYLER', 48457),
        ('UPSHUR', 48459),
        ('UPTON', 48461),
        ('UVALDE', 48463),
        ('VAL VERDE', 48465),
        ('VAN ZANDT', 48467),
        ('VICTORIA', 48469),
        ('WALKER', 48471),
        ('WALLER', 48473),
        ('WARD', 48475),
        ('WASHINGTON', 48477),
        ('WEBB', 48479),
        ('WHARTON', 48481),
        ('WHEELER', 48483),
        ('WICHITA', 48485),
        ('WILBARGER', 48487),
        ('WILLACY', 48489),
        ('WILLIAMSON', 48491),
        ('WILSON', 48493),
        ('WINKLER', 48495),
        ('WISE', 48497),
        ('WOOD', 48499),
        ('YOAKUM', 48501),
        ('YOUNG', 48503),
        ('ZAPATA', 48505),
        ('ZAVALA', 48507),
    )

    df = pandas.DataFrame(np.array(codes))
    df = df.rename(columns={0: 'name', 1: 'fips'})
    df['fips'] = df['fips'].astype(int)
    return df


def _get_text_url(date):
    return 'http://twc.tamu.edu/weather_images/summ/summ%s.txt' % date.strftime('%Y%m%d')

def _get_csv_url(date):
    return 'http://twc.tamu.edu/weather_images/summ/summ%s.csv' % date.strftime('%Y%m%d')

def _parse_text_file(data_file):
    """
    example:
        COUNTY                        KBDI_AVG   KBDI_MAX    KBDI_MIN
                ----------------------------------------------------------------
                ANDERSON                         262       485        47
                ANDREWS                          485       614       357
                ...
    """

    dtype = [
        ('county', '|U15'),
        ('avg', 'i4'),
        ('max', 'i4'),
        ('min', 'i4'),
    ]

    if not data_file.readline().lower().startswith(b'county'):
        return pandas.DataFrame()
    data_file.seek(0)

    data_array = np.genfromtxt(
        data_file, delimiter=[31, 11, 11, 11], dtype=dtype, skip_header=2,
        skip_footer=1, autostrip=True)
    dataframe = pandas.DataFrame(data_array)
    return dataframe

def _parse_csv_file(data_file):
    """
    example:
        County,Min,Max,Average,Change
        Anderson,429,684,559,+5
        Andrews,92,356,168,+7
    """

    if not data_file.readline().lower().startswith(b'county'):
        return pandas.DataFrame()
    data_file.seek(0)

    dataframe = pandas.read_csv(data_file)
    dataframe.columns = dataframe.columns.str.lower()
    dataframe = dataframe.rename(columns={'average':'avg'})
    dataframe.county = dataframe.county.str.upper()
    dataframe = dataframe[['county','avg','max','min']]

    return dataframe

def _open_data_file(url, data_dir):
    """returns an open file handle for a data file; downloading if necessary or
    otherwise using a previously downloaded file
    """
    file_name = url.rsplit('/', 1)[-1]
    file_path = os.path.join(data_dir, file_name)
    return util.open_file_for_url(url, file_path, check_modified=True, use_bytes=True)



"""
    Access to data provided by the `United States Army
    Corps of Engineers`_ - `Tulsa District Water Control`_ web site.

    .. _United States Army Corps of Engineers: http://www.usace.army.mil/
    .. _Tulsa District Water Control: http://www.swt-wc.usace.army.mil/
"""
from .core import get_station_data, get_stations



"""
    `NASA EARTHDATA ORNL DAAC Daymet`_ web services


    .. _NASA EARTHDATA ORNL DAAC Daymet: https://daymet.ornl.gov/dataaccess.html
"""
from __future__ import absolute_import

from . import core

from .core import (get_daymet_singlepixel)
from .core import (get_variables)

from ulmo import util


import ulmo

import test_util


def test_parse_site_infos():
    site_files = ['usgs/nwis/RI_daily.xml', 'usgs/nwis/RI_instantaneous.xml']
    site_infos = {}
    for site_file in site_files:
        test_site_file = test_util.get_test_file_path(site_file)
        with open(test_site_file, 'rb') as f:
            site_infos.update(ulmo.waterml.v1_1.parse_site_infos(f))

    assert len(site_infos) == 64

    test_sites = {
        '01111410': {
            'agency': 'USGS',
            'code': '01111410',
            'location': {
                'latitude': '41.9409318',
                'longitude': '-71.6481214',
                'srs': 'EPSG:4326',
            },
            'name': 'CHEPACHET RIVER WEST OF GAZZA RD AT GAZZAVILLE, RI',
            'network': 'NWIS',
            'timezone_info': {
                'default_tz': {
                    'abbreviation': 'EST',
                    'offset': '-05:00'
                },
                'dst_tz': {
                    'abbreviation': 'EDT',
                    'offset': '-04:00',
                },
                'uses_dst': False,
            },
            'site_property': {
                'site_type_cd': 'ST',
                'huc_cd': '01090003',
                'state_cd': '44',
                'county_cd': '44007',
            }
        }
    }

    for test_code, test_dict in test_sites.items():
        assert site_infos[test_code] == test_dict
    return site_infos


def test_parse_site_values():
    query_isodate = '2000-01-01'
    value_file = test_util.get_test_file_path(
            'usgs/nwis/site_07335390_instantaneous.xml')
    with open(value_file, 'rb') as content_io:
        values = ulmo.waterml.v1_1.parse_site_values(content_io, query_isodate)

    assert len(values['00062:00011']['values']) > 1000



"""
    ulmo.cpc.drought.core
    ~~~~~~~~~~~~~~~~~~~~~

    This module provides direct access to `Climate Predicition Center`_ `Weekly
    Drought Index`_ dataset.

    .. _Climate Prediction Center: http://www.cpc.ncep.noaa.gov/
    .. _Weekly Drought Index: http://www.cpc.ncep.noaa.gov/products/analysis_monitoring/cdus/palmer_drought/
"""
from __future__ import division
from builtins import str
from builtins import range
from past.utils import old_div

import datetime
import os
import requests

import numpy as np
import pandas

from ulmo import util

# directory where drought data will be stashed
CPC_DROUGHT_DIR = os.path.join(util.get_ulmo_dir(), 'cpc/drought')

# state codes (note: these are not FIPS codes)
STATE_CODES = {
    'AL': 1,
    'AZ': 2,
    'AR': 3,
    'CA': 4,
    'CO': 5,
    'CT': 6,
    'DE': 7,
    'FL': 8,
    'GA': 9,
    'IA': 13,
    'ID': 10,
    'IL': 11,
    'IN': 12,
    'KS': 14,
    'KY': 15,
    'LA': 16,
    'MA': 19,
    'MD': 18,
    'ME': 17,
    'MI': 20,
    'MN': 21,
    'MO': 23,
    'MS': 22,
    'MT': 24,
    'NC': 31,
    'ND': 32,
    'NE': 25,
    'NH': 27,
    'NJ': 28,
    'NM': 29,
    'NV': 26,
    'NY': 30,
    'OH': 33,
    'OK': 34,
    'OR': 35,
    'PA': 36,
    'PR': 66,
    'RI': 37,
    'SC': 38,
    'SD': 39,
    'TN': 40,
    'TX': 41,
    'UT': 42,
    'VA': 44,
    'VT': 43,
    'WA': 45,
    'WI': 47,
    'WV': 46,
    'WY': 48,
}

def get_data(state=None, climate_division=None, start=None, end=None,
             as_dataframe=False):
    """Retreives data.


    Parameters
    ----------
    state : ``None`` or str
        If specified, results will be limited to the state corresponding to the
        given 2-character state code.
    climate_division : ``None`` or int
        If specified, results will be limited to the climate division.
    start : ``None`` or date (see :ref:`dates-and-times`)
        Results will be limited to those after the given date. Default is the
        start of the current calendar year.
    end : ``None`` or date (see :ref:`dates-and-times`)
        If specified, results will be limited to data before this date.
    as_dataframe: bool
        If ``False`` (default), a dict with a nested set of dicts will be
        returned with data indexed by state, then climate division. If ``True``
        then a pandas.DataFrame object will be returned.  The pandas dataframe
        is used internally, so setting this to ``True`` is a little bit faster
        as it skips a serialization step.


    Returns
    -------
    data : dict or pandas.Dataframe
        A dict or pandas.DataFrame representing the data. See the
        ``as_dataframe`` parameter for more.
    """
    if not start is None:
        start_date = util.convert_date(start)
    else:
        start_date = None
    if not end is None:
        end_date = util.convert_date(end)
    else:
        end_date = None

    if not end_date:
        end_date = datetime.date.today()
    if not start_date:
        start_date = datetime.date(end_date.year, 1, 1)

    start_year, start_week = _week_number(start_date)
    end_year, end_week = _week_number(end_date)

    if state:
        state_code = STATE_CODES.get(state.upper())
    else:
        state_code = None

    data = None
    for year in range(start_year, end_year + 1):
        url, current_year_flag = _get_data_url(year)
        format_type = _get_data_format(year)
        with _open_data_file(url) as data_file:
            year_data = _parse_data_file(data_file, format_type, year, current_year_flag)

        if state_code:
            year_data = year_data[year_data['state_code'] == state_code]
        if climate_division:
            year_data = year_data[year_data['climate_division'] == climate_division]

        year_data = _reindex_data(year_data)

        if data is None:
            data = year_data
        else:
            # some data are duplicated (e.g. final data from 2011 stretches into
            # prelim data of 2012), so just take those that are new
            append_index = year_data.index.difference(data.index)
            if len(append_index):
                data = data.append(year_data.loc[append_index])

    # restrict results to date range
    period_index = pandas.PeriodIndex(data['period'])
    periods_in_range = (period_index >= start_date) & (period_index <= end_date)
    data = data[periods_in_range]

    # this does what data.reset_index() should do, but at least as of 0.10.1, that sets
    # will cast period objects to ints
    data.index = np.arange(len(data))
    if as_dataframe:
        return data
    else:
        return _as_data_dict(data)


def _as_data_dict(dataframe):
    data_dict = {}
    for state in dataframe['state'].unique():
        state_dict = {}
        state_dataframe = dataframe[dataframe['state'] == state]
        for name, group in state_dataframe.groupby(['state', 'climate_division']):
            s, climate_division = name
            climate_division_data = group.T.drop(['state', 'climate_division'])
            values = [
                _value_dict(value)
                for k, value in climate_division_data.iteritems()
            ]
            state_dict[climate_division] = values
        data_dict[state] = state_dict
    return data_dict


def _convert_state_codes(dataframe):
    """adds state abbreviations to a dataframe, based on state codes"""
    state_codes = pandas.DataFrame(
        np.array([i for i in STATE_CODES.items()],
                 dtype=np.dtype([('state', '|U2'), ('code', int)])))
    merged = pandas.merge(dataframe, state_codes,
            left_on='state_code', right_on='code', how='left')
    column_names = dataframe.columns.tolist()
    column_names.remove('state_code')
    column_names.insert(0, 'state')
    return merged[column_names]


def _convert_week_numbers(dataframe):
    """convert a dataframe's week numbers to period objects"""
    weeks = [key for key, group in dataframe.groupby(['year', 'week'])]
    periods = [(week[0], week[1], _period_for_week(*week)) for week in weeks]
    period_dataframe = pandas.DataFrame(periods, columns=['year', 'week', 'period'])
    merged = pandas.merge(dataframe, period_dataframe,
            left_on=['year', 'week'], right_on=['year', 'week'])
    column_names = dataframe.columns.tolist()
    column_names.remove('week')
    column_names.remove('year')
    column_names.insert(2, 'period')
    return merged[column_names]


def _first_sunday(year):
    """returns the first Sunday of a growing season, which is the first Sunday
    after the first Wednesday in March
    """
    first_day = datetime.date(year, 3, 1)
    if first_day.weekday() == 6:
        return first_day
    elif first_day.weekday() <= 2:
        return first_day - pandas.tseries.offsets.Week(weekday=6)
    else:
        return first_day + pandas.tseries.offsets.Week(weekday=6)


def _get_data_format(year):
    if year >= 2001:
        return 'format5'
    elif 1997 <= year <= 2000:
        return 'format4'
    else:
        return 'format2'


def _get_data_url(year):
    current_year, current_week = _week_number(datetime.date.today())
    if year == current_year:
        return ('https://ftp.cpc.ncep.noaa.gov/htdocs/temp4/current.data', True)
    elif year == current_year - 1:
        url = ('https://ftp.cpc.ncep.noaa.gov/htdocs/temp2/palmer%s-PRELIM' % str(year)[-2:],
                False)
        if not _url_exists(url[0]):
            url = ('https://ftp.cpc.ncep.noaa.gov/htdocs/temp4/current.data', True)
        return url
    elif year <= 1985:
        return ('https://ftp.cpc.ncep.noaa.gov/htdocs/temp2/palmer73-85', False)
    else:
        url = ('https://ftp.cpc.ncep.noaa.gov/htdocs/temp2/palmer%s' % str(year)[-2:], False)
        if not _url_exists(url[0]):
            url = ('https://ftp.cpc.ncep.noaa.gov/htdocs/temp2/palmer%s-PRELIM' % str(year)[-2:],
                    False)
        return url


def _open_data_file(url):
    """returns an open file handle for a data file; downloading if necessary or otherwise using a previously downloaded file"""
    file_name = url.rsplit('/', 1)[-1]
    file_path = os.path.join(CPC_DROUGHT_DIR, file_name)
    return util.open_file_for_url(url, file_path, check_modified=True, use_bytes=True)


def _parse_data_file(data_file, palmer_format, year, current_year_flag):
    """
    based on the fortran format strings:
        format2: FORMAT(I4,3I2,F4.1,F4.0,10F6.2,4F6.4,F6.3,10F6.2,F4.0,12F6.2)
        format4: FORMAT(2I4,I2,F4.1,F4.0,10F6.2,4F6.4,F6.3,10F6.2,F4.0,12F6.2)
        format5: FORMAT(2I4,I2,F5.2,F5.1,10F6.2,4F6.4,F6.3,10F6.2,F4.0,12F6.2)
    """
    if palmer_format == 'format5':
        delim_sequence = (2, 2, 4, 2, 5, 5) + 10*(6,) + 4*(6,) + (6,) + 10*(6,) + (4,) + 12*(6,)
        use_columns = (0, 1, 2, 3, 4, 5, 9, 15, 28, 29, 37, 40, 41)
    elif palmer_format == 'format4':
        delim_sequence = (2, 2, 4, 2, 4, 4) + 10*(6,) + 4*(6,) + (6,) + 10*(6,) + (4,) + 12*(6,)
        use_columns = (0, 1, 2, 3, 4, 5, 9, 15, 28, 29, 37, 40, 41)
    elif palmer_format == 'format2':
        delim_sequence = (2, 2, 2, 2, 2, 4, 4) + 10*(6,) + 4*(6,) + (6,) + 10*(6,) + (4,) + 12*(6,)
        use_columns = (0, 1, 2, 3, 5, 6, 10, 16, 29, 30, 38, 41, 42)
    else:
        raise NotImplementedError("we have not implemented the format for given date range")

    dtype = [
        ('state_code', 'i1'),
        ('climate_division', 'i1'),
        ('year', 'i4'),
        ('week', 'i4'),
        ('precipitation', 'f8'),
        ('temperature', 'f8'),
        ('potential_evap', 'f8'),
        ('runoff', 'f8'),
        ('soil_moisture_upper', 'f8'),
        ('soil_moisture_lower', 'f8'),
        ('pdsi', 'f8'),
        ('cmi', 'f8')
    ]

    decodef = lambda x: x.decode("utf-8")
    data_array = np.genfromtxt(data_file, dtype=dtype, delimiter=delim_sequence, usecols=use_columns)
    if not current_year_flag:
        data_array['year'] = year
    dataframe = pandas.DataFrame(data_array)
    return dataframe


def _periods_for_range(start_date, end_date):
    return pandas.period_range(start_date, end_date, freq='W-SAT')


def _period_for_week(year, week_number):
    """returns a pandas.Period for a given growing season year and week number"""
    first_sunday = _first_sunday(year)
    return pandas.Period(first_sunday, freq='W-SAT') + week_number - 1


def _reindex_data(dataframe):
    dataframe = _convert_week_numbers(dataframe)
    dataframe = _convert_state_codes(dataframe)
    return dataframe.set_index(
        ['state', 'climate_division', 'period'], drop=False)


def _url_exists(url):
    return requests.head(url).status_code == 200


def _value_dict(value):
    value_dict = value.to_dict()
    value_dict['period'] = str(value_dict['period'])
    return value_dict


def _week_number(date):
    """returns the growing season week number for a given datetime.date"""
    first_sunday = _first_sunday(date.year)
    date_ts = pandas.Timestamp(date)
    first_sunday_ts = pandas.Timestamp(first_sunday)
    if date_ts < first_sunday_ts:
        first_sunday_ts = pandas.Timestamp(_first_sunday(date.year - 1))
    days_since_first_sunday = (date_ts - first_sunday_ts).days
    return (first_sunday_ts.year, (old_div(days_since_first_sunday, 7)) + 1)



"""
    ulmo.usace.swtwc.core
    ~~~~~~~~~~~~~~~~~~~~~

    This module provides access to data provided by the `United States Army
    Corps of Engineers`_ `Tulsa District Water Control`_ web site.

    .. _United States Army Corps of Engineers: http://www.usace.army.mil/
    .. _Tulsa District Water Control: http://www.swt-wc.usace.army.mil/

"""
from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import zip
from builtins import range
import datetime
import os.path

from bs4 import BeautifulSoup
import numpy as np
import requests
import pandas

from ulmo import util

try:
    import io as StringIO
except ImportError:
    from cStringIO import StringIO

USACE_SWTWC_DIR = os.path.join(util.get_ulmo_dir(), 'usace/swtwc')


def get_station_data(station_code, date=None, as_dataframe=False):
    """Fetches data for a station at a given date.


    Parameters
    ----------
    station_code: str
        The station code to fetch data for. A list of stations can be retrieved with
        ``get_stations()``
    date : ``None`` or date (see :ref:`dates-and-times`)
        The date of the data to be queried. If date is ``None`` (default), then
        data for the current day is retreived.
    as_dataframe : bool
        This determines what format values are returned as. If ``False``
        (default), the values dict will be a dict with timestamps as keys mapped
        to a dict of gauge variables and values. If ``True`` then the values
        dict will be a pandas.DataFrame object containing the equivalent
        information.


    Returns
    -------
    data_dict : dict
        A dict containing station information and values.
    """

    station_dict = {}
    if date is None:
        date_str = 'current'
        year = datetime.date.today().year
    else:
        date = util.convert_date(date)
        date_str = date.strftime('%Y%m%d')
        year = date.year

    filename = '{}.{}.html'.format(station_code, date_str)
    data_url = 'http://www.swt-wc.usace.army.mil/webdata/gagedata/' + filename

    # requests without User-Agent header get rejected
    headers = {
        'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/72.0.3626.121 Safari/537.36'
    }
    resp = requests.get(data_url, headers=headers)
    soup = BeautifulSoup(resp.content)
    pre = soup.find('pre')
    if pre is None:
        error_msg = 'no data could be found for station code %(station_code)s and date %(date)s (url: %(data_url)s)' % {
            'date': date,
            'data_url': data_url,
            'station_code': station_code,
        }
        raise ValueError(error_msg)
    sio = StringIO.StringIO(str(pre.text.strip()))

    first_line = sio.readline()
    split = first_line[8:].strip().split()

    station_dict['code'] = split[0]
    station_dict['description'] = ' '.join(split[1:])

    second_line = sio.readline()
    station_dict['station_type'] = second_line.strip().split(':')[1].strip()

    notes = []

    while 1:
        next_line = sio.readline()
        if ':' in next_line:
            notes.append(next_line.strip())
        else:
            break

    if len(notes):
        station_dict['notes'] = '\n'.join(notes)

    variable_names = _split_line(sio.readline()[11:], 10)
    variable_units = _split_line(sio.readline()[11:], 10)
    variable_sources = _split_line(sio.readline()[11:], 10)

    station_dict['variables'] = dict([
        (name, {'unit': unit, 'source': source})
        for name, unit, source in zip(
            variable_names, variable_units, variable_sources)
    ])

    station_dict['timezone'] = sio.readline().strip().strip('()')
    column_names = ['datetime'] + variable_names
    widths = [14] + ([10] * len(variable_names))
    converters = dict([
        (variable_name, lambda x: float(x) if x != '----' else np.nan)
        for variable_name in variable_names
    ])
    date_parser = lambda x: _convert_datetime(x, year)
    dataframe = pandas.read_fwf(
        sio, names=column_names, widths=widths, index_col=['datetime'],
        na_values=['----'], converters=converters, parse_dates=True,
        date_parser=date_parser)

    # parse out rows that are all nans (e.g. end of "current" page)
    dataframe = dataframe[~np.isnan(dataframe.T.sum())]

    if as_dataframe:
        station_dict['values'] = dataframe
    else:
        station_dict['values'] = util.dict_from_dataframe(dataframe)

    return station_dict


def get_stations():
    """Fetches a list of station codes and descriptions.

    Returns
    -------
    stations_dict : dict
        a python dict with station codes mapped to station information
    """
    stations_url = 'http://www.swt-wc.usace.army.mil/shefids.htm'

    headers = {
        'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/72.0.3626.121 Safari/537.36'
    }
    resp = requests.get(stations_url, headers=headers)
    soup = BeautifulSoup(resp.content)
    pre = soup.find('pre')
    links = pre.find_all('a')
    stations = [
        _parse_station_link(link) for link in links
    ]

    return dict([
        (station['code'], station)
        for station in stations
    ])


def _convert_datetime(s, year):
    fmt = '%m/%d %H:%M'
    return datetime.datetime.strptime(s, fmt).replace(year=year)


def _split_line(line, n):
    return [line[i:i + n].strip() for i in range(0, len(line), n)][:-1]


def _parse_station_link(link):
    return {
        'code': link.text,
        'description': link.next_sibling.strip(),
    }


def _to_underscore(spaced):
    return spaced.sub(' ', '_').sub('(', '').sub(')', '').lower()



import pytest
import ulmo
import test_util
import filecmp


test_sets = [
	{'layer': '1 arc-second',
	 'bbox': (-79.68433821243661, 32.81417227179455, -77.42742156945509, 34.798064728936754),
	 'number_of_tiles': 7,
	 'file': 'usgs/ned/availability_bbox_test_set_1.json',
	},
	{'layer': '1/3 arc-second',
	 'bbox': (-79.68433821243661, 32.81417227179455, -77.42742156945509, 34.798064728936754),
	 'number_of_tiles': 7,
	 'file': 'usgs/ned/availability_bbox_test_set_2.json',
	},
	{'layer': 'Alaska 2 arc-second',
	 'bbox': (-150.4969, 60.7338, -148.4600, 61.5253),
	 'number_of_tiles': 6,
	 'file': 'usgs/ned/availability_bbox_test_set_3.json',
	},
]


def test_get_raster_availability():
	for dataset in test_sets:
		with test_util.mocked_urls(dataset['file']):
			locs = ulmo.usgs.ned.get_raster_availability(dataset['layer'], dataset['bbox'])
			assert len(locs['features'])==dataset['number_of_tiles']


def test_get_raster():
	layer = 'Alaska 2 arc-second'
	bbox = (-149.5, 60.5, -149.3, 60.7)
	#availability_url = 'https://www.sciencebase.gov/catalog/items?fields=webLinks,spatial,title&q=&filter=tags=National Elevation Dataset (NED) Alaska 2 arc-second&filter=tags=IMG&filter=spatialQuery=Polygon ((-149.5 60.7,-149.5 60.5,-149.3 60.5,-149.3 60.7,-149.5 60.7))&format=json&max=1000'
	#zip_url = 'ftp://rockyftp.cr.usgs.gov/vdelivery/Datasets/Staged/NED/2/IMG/n61w150.zip'
	availability_url = 'https://www.sciencebase.gov/catalog/*'
	zip_url = 'ftp://rockyftp.cr.usgs.gov/vdelivery/Datasets/*'

	url_files = {
		availability_url: 'usgs/ned/get_raster_test_availability.json',
		zip_url: 'n61w150.zip',
	}

	test_file = test_util.get_test_file_path('usgs/ned/USGS_NED_2_n61w150_IMG.img')
	with test_util.temp_dir() as data_dir:
		#with test_util.mocked_urls(url_files):
		locs = ulmo.usgs.ned.get_raster(layer, bbox, path=data_dir)
		raster_tile = locs['features'][0]['properties']['file']
		assert filecmp.cmp(raster_tile, test_file)



from past.builtins import basestring
import copy

import numpy as np
import pandas

import ulmo
import test_util


test_sets = [
    {
        'element': 'pdsi',
        'by_state': False,
        'values': [
            {
                'state': 'WY',
                'state_code': 48,
                'division': 8,
                'year': 1998,
                'month': 1,
                'pdsi': 1.84,
            }, {
                'state': 'WY',
                'state_code': 48,
                'division': 8,
                'year': 1998,
                'month': 2,
                'pdsi': 1.88,
            }, {
                'state': 'WY',
                'state_code': 48,
                'division': 8,
                'year': 1998,
                'month': 3,
                'pdsi': 1.99,
            }, {
                'state': 'WY',
                'state_code': 48,
                'division': 8,
                'year': 1998,
                'month': 11,
                'pdsi': 2.71,
            }, {
                'state': 'WY',
                'state_code': 48,
                'division': 8,
                'year': 1998,
                'month': 12,
                'pdsi': 2.53,
            }, {
                'state': 'TX',
                'state_code': 41,
                'division': 10,
                'year': 1895,
                'month': 1,
                'pdsi': -0.05,
            }, {
                'state': 'TX',
                'state_code': 41,
                'division': 10,
                'year': 1895,
                'month': 12,
                'pdsi': 1.72,
            },
        ]
    }, {
        'element': 'tmpc',
        'by_state': True,
        'values': [
            {
                'location': 'ME',
                'location_code': 17,
                'year': 1895,
                'month': 1,
                'tmpc': 12.20,
            }, {
                'location': 'ME',
                'location_code': 17,
                'year': 1895,
                'month': 12,
                'tmpc': 20.7,
            }, {
                'location': 'national',
                'location_code': 110,
                'year': 2013,
                'month': 1,
                'tmpc': 32.25,
            }, {
                'location': 'national',
                'location_code': 110,
                'year': 2013,
                'month': 2,
                'tmpc': 34.79,
            },
        ]
    }, {
        'element': 'tmpc',
        'by_state': True,
        'location_names': 'full',
        'values': [
            {
                'location': 'Maine',
                'location_code': 17,
                'year': 1895,
                'month': 1,
                'tmpc': 12.20,
            },
        ]
    }, {
        'element': 'sp01',
        'by_state': True,
        'values': [
            {
                'location': 'AL',
                'location_code': 1,
                'year': 1895,
                'month': 1,
                'sp01': 1.23,
            },
        ]
    }, {
        'element': ['sp01', 'tmpc', 'pdsi'],
        'by_state': True,
        'values': [
            {
                'location': 'AL',
                'location_code': 1,
                'year': 1895,
                'month': 1,
                'sp01': 1.23,
                'tmpc': 43.1,
                'pdsi': 0.78,
            },
            {
                'location': None,
                'location_code': 113,
                'year': 1895,
                'month': 1,
                'sp01': None,
                'tmpc': 15.5,
                'pdsi': None,
            },
        ]
    },
]


def test_get_data_by_state():
    state_tests = [s for s in test_sets if s['by_state']]
    _run_test_sets(state_tests)


def test_get_data_by_climate_division():
    division_tests = [s for s in test_sets if not s['by_state']]
    _run_test_sets(division_tests)


def test_doesnt_have_locations_if_location_names_is_none():
    element = 'pdsi'
    use_file = _test_use_file(element, False)
    data = ulmo.ncdc.cirs.get_data(
        element, location_names=None, use_file=use_file, as_dataframe=True)
    assert 'location' not in data.columns


def _run_test_sets(test_sets):
    for test_set in test_sets:
        test_args = copy.copy(test_set)
        element = test_args.pop('element')
        test_values = test_args.pop('values')
        by_state = test_args['by_state']

        use_file = _test_use_file(element, by_state)
        data = ulmo.ncdc.cirs.get_data(
            element, use_file=use_file, as_dataframe=True, **test_args)
        if isinstance(element, basestring) or len(element) == 1:
            not_nulls = data[element].notnull()
        else:
            not_nulls = pandas.notnull(data[element]).any(axis=1)
        assert np.all(not_nulls)
        for test_value in test_values:
            _assert_inclusion(test_value, data)


def _assert_inclusion(value_dict, dataframe):
    "tests that a value_dict is in a dataframe"
    sub_df = dataframe.copy()
    for k, v in value_dict.items():
        if pandas.isnull(v):
            sub_df = sub_df[pandas.isnull(sub_df[k])]
        else:
            sub_df = sub_df[sub_df[k] == v]

    assert len(sub_df) == 1


def _test_use_file(elements, by_state):
    if test_util.use_test_files():
        if isinstance(elements, basestring):
            div = "st" if by_state else "dv"
            path = 'ncdc/cirs/climdiv-{elements}{div}-v1.0.0-20140304'.format(
                elements=elements, div=div)
        else:
            path = 'ncdc/cirs/'
        return test_util.get_test_file_path(path)
    else:
        return None



"""
    ulmo.twc.kbdi.core
    ~~~~~~~~~~~~~~~~~~~~~

    This module provides direct access to `Texas Weather Connection`_ -
    `Daily Keetch-Byram Drought Index (KBDI)`_ dataset.

    .. _Texas Weather Connection: http://twc.tamu.edu/
    .. _Daily Keetch-Byram Drought Index (KBDI): http://twc.tamu.edu/drought/kbdi
"""
from .core import get_data



import datetime
import pandas

import ulmo

import test_util


def test_get_sites_by_type():
    sites_file = 'lcra/hydromet/stream_stage_and_flow_sites_list.html'
    with test_util.mocked_urls(sites_file):
        sites = ulmo.lcra.hydromet.get_sites_by_type('stage')
        assert 60 <= len(sites) <= 70
        assert '5499' in sites


def test_get_site_data():
    test_values = pandas.DataFrame(
        [{'Stage(feet)': 6.20, 'Flow(cfs)': 74},
         {'Stage(feet)': 6.01, 'Flow(cfs)': 58}],
        index=[datetime.datetime(2015, 11, 28, 2, 55, 0),
               datetime.datetime(2015, 12, 3, 10, 10, 0)])
    data_file = 'lcra/hydromet/4598_stage_flow_data.html'
    with test_util.mocked_urls(data_file):
        site_data = ulmo.lcra.hydromet.get_site_data(
            '4598', 'stage', start_date=datetime.date(2015, 11, 3),
            end_date=datetime.date(2015, 12, 4))

    assert site_data.shape[0] == 2932
    are_equal = test_values == site_data.ix[test_values.index]
    assert are_equal.sum().sum() == 4


def test_get_current_data():
    test_values = pandas.DataFrame(
        [{'datetime': datetime.datetime(2015, 12, 10, 14, 10),
          'location': 'Barton Creek at Loop 360, Austin',
          'stageft': 3.33,
          'flowcfs': 60.00,
          'floodstageft': 8.00,
          'bankfullstageft': 8.00
          },
          {'datetime': datetime.datetime(2015, 12, 10, 14, 10),
           'location': 'Colorado River at Columbus',
           'stageft': 10.32,
           'flowcfs': 975.00,
           'bankfullstageft': 30.00,
           'floodstageft': 34.00}])
    test_values.set_index('location', inplace=True)
    data_file = 'lcra/hydromet/current_data_2015-12-10-14-10.xml'
    with test_util.mocked_urls(data_file):
        current_data = ulmo.lcra.hydromet.get_current_data('getlowerbasin')
        current_data_df = pandas.DataFrame(current_data)
        current_data_df.set_index('location', inplace=True)
    are_equal = test_values == current_data_df.ix[test_values.index][test_values.columns]
    assert pandas.np.all(are_equal)
    assert len(current_data) == 33



from builtins import range
import datetime
import ulmo

import test_util


def test_get_stations():
    test_stations = [
        ('999999-94645', {
            'USAF': '999999',
            'WBAN': '94645',
            'begin': datetime.date(2002, 9, 20),
            'icao': '',
            'country': 'US',
            'elevation': 224.6,
            'end': datetime.date(2015, 3, 11),
            'latitude': 46.96,
            'longitude': -67.883,
            'name': 'LIMESTONE 4 NNW',
            'state': 'ME',
        }),
        ('999999-14896', {
            'USAF': '999999',
            'WBAN': '14896',
            'begin': datetime.date(1952, 1, 1),
            'icao': '',
            'country': 'US',
            'elevation': 323.1,
            'end': datetime.date(1957, 11, 30),
            'latitude': 41.033,
            'longitude': -81.483,
            'name': 'AKRON NAS',
            'state': 'OH'
        }),
        ('014881-99999', {
            'USAF': '014881',
            'WBAN': '99999',
            'begin': datetime.date(2010, 11, 23),
            'icao': 'ENFB',
            'country': 'NO',
            'elevation': 142.0,
            'end': datetime.date(2015, 3, 11),
            'latitude': 61.206,
            'longitude': 1.829,
            'name': 'STATFJORD B OIL PLATFORM',
            'state': ''
        }),
        ('997944-99999', {
            'USAF': '997944',
            'WBAN': '99999',
            'begin': datetime.date(2008, 7, 24),
            'icao': '',
            'country': 'US',
            'elevation': None,
            'end': datetime.date(2009, 9, 27),
            'latitude': None,
            'longitude': None,
            'name': 'DRIFTING BUOY 48592',
            'state': '',
        }),
    ]
    with test_util.mocked_urls('ncdc/gsod/isd-history.csv'):
        stations = ulmo.ncdc.gsod.get_stations()

    assert 27000 < len(stations) < 32000

    for test_code, test_station in test_stations:
        # drop end date from test
        del stations[test_code]['end']
        del test_station['end']
        assert stations[test_code] == test_station


def test_get_stations_with_country():
    with test_util.mocked_urls('ncdc/gsod/isd-history.csv'):
        stations = ulmo.ncdc.gsod.get_stations(country='AF')
        assert len(stations) > 100
        assert '409900-99999' in stations

        stations = ulmo.ncdc.gsod.get_stations(country=['US', 'MX', 'CA'])
        # Live service now (2021-8-30) returns 9437 stations
        # assert len(stations) > 9500
        assert len(stations) > 9400
        assert '768420-99999' in stations
        assert '911660-99999' in stations
        assert '729675-99999' in stations


def test_get_stations_with_state():
    with test_util.mocked_urls('ncdc/gsod/isd-history.csv'):
        stations = ulmo.ncdc.gsod.get_stations(state='TX')
        assert len(stations) > 450
        assert '999999-93987' in stations

        stations = ulmo.ncdc.gsod.get_stations(state=['TX', 'AR', 'LA'])
        assert len(stations) > 750
        assert '999999-93987' in stations
        assert '999999-13963' in stations
        assert '994780-99999' in stations


def test_get_stations_with_start():
    with test_util.mocked_urls('ncdc/gsod/isd-history.csv'):
        stations = ulmo.ncdc.gsod.get_stations(start='2011-3-2')
    assert len(stations) > 15000
    assert '062390-99999' in stations
    assert '534780-99999' not in stations


def test_get_stations_with_end():
    with test_util.mocked_urls('ncdc/gsod/isd-history.csv'):
        stations = ulmo.ncdc.gsod.get_stations(end='1960-11-5')
    assert len(stations) > 8500
    assert '534780-99999' in stations
    assert '062390-99999' not in stations


def test_get_station_data():
    test_data = [
        (dict(
            station_codes='999999-14896',
            start='1952-01-01',
            end='1953-02-02'),
            [{
                'FRSHTT': '010000',
                'USAF': '999999',
                'WBAN': '14896',
                'date': datetime.date(1952, 1, 1),
                'dew_point': 52.4,
                'dew_point_count': 8,
                'max_gust': 999.9,
                'max_temp': 62.1,
                'max_temp_flag': '*',
                'max_wind_speed': 12.0,
                'mean_temp': 56.0,
                'mean_temp_count': 8,
                'mean_wind_speed': 9.0,
                'mean_wind_speed_count': 8,
                'min_temp': 48.0,
                'min_temp_flag': '*',
                'precip': 99.99,
                'precip_flag': ' ',
                'sea_level_pressure': 1014.4,
                'sea_level_pressure_count': 8,
                'snow_depth': 999.9,
                'station_pressure': 978.2,
                'station_pressure_count': 8,
                'visibility': 5.5,
                'visibility_count': 8,
            }, {
                'FRSHTT': '000000',
                'USAF': '999999',
                'WBAN': '14896',
                'date': datetime.date(1952, 1, 3),
                'dew_point': 18.4,
                'dew_point_count': 9,
                'max_gust': 999.9,
                'max_temp': 30.9,
                'max_temp_flag': '*',
                'max_wind_speed': 10.1,
                'mean_temp': 27.2,
                'mean_temp_count': 9,
                'mean_wind_speed': 8.3,
                'mean_wind_speed_count': 9,
                'min_temp': 23.0,
                'min_temp_flag': '*',
                'precip': 0.0,
                'precip_flag': 'I',
                'sea_level_pressure': 1033.0,
                'sea_level_pressure_count': 9,
                'snow_depth': 999.9,
                'station_pressure': 991.4,
                'station_pressure_count': 9,
                'visibility': 10.6,
                'visibility_count': 9,
            }, {
                'FRSHTT': '001000',
                'USAF': '999999',
                'WBAN': '14896',
                'date': datetime.date(1953, 2, 2),
                'dew_point': 12.0,
                'dew_point_count': 20,
                'max_gust': 999.9,
                'max_temp': 36.0,
                'max_temp_flag': '*',
                'max_wind_speed': 13.0,
                'mean_temp': 22.6,
                'mean_temp_count': 20,
                'mean_wind_speed': 8.2,
                'mean_wind_speed_count': 20,
                'min_temp': 14.0,
                'min_temp_flag': '*',
                'precip': 99.99,
                'precip_flag': ' ',
                'sea_level_pressure': 1022.8,
                'sea_level_pressure_count': 20,
                'snow_depth': 999.9,
                'station_pressure': 982.6,
                'station_pressure_count': 20,
                'visibility': 9.2,
                'visibility_count': 20,
            },
            ]
         ),
        (dict(
            station_codes='999999-14896',
            start='1952-01-01',
            end='1953-02-02',
            parameters='snow_depth'),
            [{
                'date': datetime.date(1952, 1, 1),
                'snow_depth': 999.9,
            }, {
                'date': datetime.date(1952, 1, 3),
                'snow_depth': 999.9,
            }, {
                'date': datetime.date(1953, 2, 2),
                'snow_depth': 999.9,
            },
            ]
         ),
        (dict(
            station_codes='999999-14896',
            start='1952-01-01',
            end='1953-02-02',
            parameters=['station_pressure', 'dew_point']),
            [{
                'date': datetime.date(1952, 1, 1),
                'dew_point': 52.4,
                'station_pressure': 978.2,
            }, {
                'date': datetime.date(1952, 1, 3),
                'dew_point': 18.4,
                'station_pressure': 991.4,
            }, {
                'date': datetime.date(1953, 2, 2),
                'dew_point': 12.0,
                'station_pressure': 982.6,
            },
            ]
         ),

    ]

    for kwargs, test_values in test_data:
        start = kwargs.get('start')
        end = kwargs.get('end')
        station_code = kwargs.get('station_codes')

        start_year = int(start.split('-')[0])
        end_year = int(end.split('-')[0])

        url_files = dict([
            ('http://www1.ncdc.noaa.gov/pub/data/gsod/%s/gsod_%s.tar' % (year, year),
                'ncdc/gsod/gsod_%s.tar' % year)
            for year in range(start_year, end_year + 1)
        ])

        with test_util.mocked_urls(url_files):
            station_data = ulmo.ncdc.gsod.get_data(**kwargs)

        for test_value in test_values:
            assert test_value in station_data[station_code]



import numpy as np
import pandas

from ulmo import util
from ulmo.ncdc import ghcn_daily

import test_util


test_stations = [
    {
        'country': 'US',
        'elevation': 286.5,
        'gsn_flag': 'GSN',
        'hcn_flag': 'HCN',
        'id': 'USW00003870',
        'latitude': 34.8831,
        'longitude': -82.2203,
        'name': 'GREER',
        'network': 'W',
        'network_id': '00003870',
        'state': 'SC',
        'wm_oid': '72312'
    }, {
        'country': 'US',
        'elevation': 91.1,
        'gsn_flag': None,
        'hcn_flag': None,
        'id': 'USW00003850',
        'latitude': 31.2667,
        'longitude': -85.7167,
        'name': 'CAIRNS FLD FT RUCKER',
        'network': 'W',
        'network_id': '00003850',
        'state': 'AL',
        'wm_oid': None
    },
]

test_data = {
    'USW00003870': {
        'ACMH': {
            '1965-03-27': (50, np.nan, np.nan, 'X'),
        },
        'PRCP': {
            '1962-10-01': (np.nan, np.nan, np.nan, np.nan),
            '1962-10-14': (np.nan, np.nan, np.nan, np.nan),
            '1962-10-15': (0, np.nan, np.nan, 'X'),
            '1962-10-16': (0, 'T', np.nan, 'X'),
            '1962-10-17': (0, np.nan, np.nan, 'X'),
            '1962-10-21': (102, np.nan, np.nan, 'X'),
            '1962-10-31': (30, np.nan, np.nan, 'X'),
            '1963-02-01': (15, np.nan, np.nan, '0'),
            '1963-02-28': (0, np.nan, np.nan, '0'),
            '1963-03-01': (249, np.nan, np.nan, '0'),
            '1984-02-01': (0, np.nan, np.nan, '0'),
            '1984-02-06': (8, np.nan, np.nan, '0'),
            '1984-02-28': (20, np.nan, np.nan, '0'),
            '1984-02-29': (0, np.nan, np.nan, '0'),
            '1984-03-01': (0, np.nan, np.nan, '0'),
            '1984-03-05': (287, np.nan, np.nan, '0'),
            '2012-03-01': (0, 'T', np.nan, 'X'),
            '2012-03-02': (10, np.nan, np.nan, 'X'),
            '2012-03-03': (422, np.nan, np.nan, 'X'),
        },
    },
    'USC00411885': {
        'PRCP': {
            '1912-09-01': (0, 'P', np.nan, '6'),
            '1912-09-05': (0, 'P', np.nan, '6'),
            '1912-09-06': (0, 'P', np.nan, '6'),
            '1912-09-07': (0, 'P', np.nan, '6'),
            '1912-09-08': (0, 'P', np.nan, '6'),
            '1912-09-30': (0, 'P', np.nan, '6'),
        },
    },
}


def test_get_data_as_dataframes():
    for station_id, sample_data in test_data.items():
        elements = list(sample_data.keys())
        with test_util.mocked_urls('ncdc/ghcnd/%s.dly' % station_id):
            station_data = ghcn_daily.get_data(
                station_id, elements=elements, as_dataframe=True)

            for element_id, element_test_data in sample_data.items():
                element_df = station_data[element_id]
                for date, test_value in element_test_data.items():
                    date = pandas.Period(pandas.Timestamp(date), freq='D')
                    value = element_df.xs(date)
                    test_array = np.array(test_value, dtype=value.dtype)

                    nulls = pandas.isnull(value)

                    assert np.all(pandas.isnull(test_array) == nulls)
                    assert np.all(value[~nulls.values] == test_array[~nulls.values])


def test_get_data_as_dicts():
    for station_id, sample_data in test_data.items():
        elements = list(sample_data.keys())
        with test_util.mocked_urls('ncdc/ghcnd/%s.dly' % station_id):
            station_data = ghcn_daily.get_data(station_id, elements=elements)

            for element_id, element_test_data in sample_data.items():
                element_dict = station_data[element_id]
                for date, test_value in element_test_data.items():
                    value_dict = element_dict[date]
                    values = [
                        value_dict.get(v)
                        for v in ['value', 'mflag', 'qflag', 'sflag']
                    ]
                    value = np.array(values, dtype=object)
                    test_array = np.array(test_value, dtype=value.dtype)

                    nulls = pandas.isnull(value)

                    assert np.all(pandas.isnull(test_array) == nulls)
                    assert np.all(value[~nulls] == test_array[~nulls])


def test_get_stations_as_dicts():
    with test_util.mocked_urls('ncdc/ghcnd/ghcnd-stations.txt'):
        stations = ghcn_daily.get_stations()
    assert len(stations) > 80000

    for test_station in test_stations:
        station_id = test_station.get('id')
        assert stations.get(station_id) == test_station


def test_get_stations_as_dataframe():
    with test_util.mocked_urls('ncdc/ghcnd/ghcnd-stations.txt'):
        stations = ghcn_daily.get_stations(as_dataframe=True)
    assert len(stations) > 80000

    for test_station in test_stations:
        station_id = test_station.get('id')
        station = stations.xs(station_id)
        station_dict = util.misc._nans_to_nones(station.to_dict())
        assert station_dict == test_station


def test_get_stations_by_country():
    with test_util.mocked_urls('ncdc/ghcnd/ghcnd-stations.txt'):
        stations = ghcn_daily.get_stations(country='US', as_dataframe=True)
    assert len(stations) > 45000


def test_get_stations_by_state():
    with test_util.mocked_urls('ncdc/ghcnd/ghcnd-stations.txt'):
        stations = ghcn_daily.get_stations(state='TX', as_dataframe=True)
    assert len(stations) > 3200


def test_get_stations_with_date_range():
    test_ranges = [
        {
            'start': 2011,
            'end': None,
            'includes': [
                'KE000063661',
                'USS0021E07S',
                'ZI000067983',
            ],
            'excludes': [
                'JQW00021601',
                'USC00448257',
                'ZI000067991',
            ],
        }, {
            'start': None,
            'end': 1920,
            'includes': [
                'USW00094957',
                'WZ004451000',
                'WZ004455110',
                'WZ004834260',
            ],
            'excludes': [
                'ASN00005063',
                'CA00405DJDN',
                'CA005040896',
            ],
        }, {
            'start': 1937,
            'end': 1945,
            'includes': [
                'ASN00015621',
                'ASN00023053',
                'WZ004451000',
                'WZ004834260',
            ],
            'excludes': [
                'ASN00041460',
                'ASN00023301',
                'US1COLR0770',
                'US1COLR0850',
            ],
        }, {
            'start': 1960,
            'end': 1960,
            'includes': [
                'ASN00078019',
                'ASN00041435',
            ],
            'excludes': [
                'US1COLR0770',
                'US1COLR0850',
            ],
        },
    ]

    url_files = {
        'ghcnd-stations': 'ncdc/ghcnd/ghcnd-stations.txt',
        'ghcnd-inventory': 'ncdc/ghcnd/ghcnd-inventory.txt'
    }

    with test_util.mocked_urls(url_files):
        for test_range in test_ranges:
            start = test_range.get('start')
            end = test_range.get('end')
            stations = ghcn_daily.get_stations(start_year=start,
                end_year=end, as_dataframe=True)
            _check_stations_dataframe(stations,
                test_range.get('includes'),
                test_range.get('excludes'))


def test_get_stations_with_elements():
    test_elements = [
        {
            'elements': 'PRCP',
            'includes': [
                'ASN00008230',
                'WA006567710',
                'VQC00672823',
            ],
            'excludes': ['SWE00136141', 'USR0000OHOR', 'USC00450935', 'USR0000THEN', 'USR0000MDRY'],
        }, {
            'elements': ['SNOW', 'TMAX'],
            'includes': [
                'ACW00011604',
                'USW00094895',
                'VQW00011640',
                'ZI000067991',
            ],
            'excludes': ['BR037642870', 'BR00C8-0100', 'BR048519530', 'BR002548070', 'IN009081600']
        },
    ]
    url_files = {
        'ghcnd-stations': 'ncdc/ghcnd/ghcnd-stations.txt',
        'ghcnd-inventory': 'ncdc/ghcnd/ghcnd-inventory.txt'
    }

    with test_util.mocked_urls(url_files):
        for test_element in test_elements:
            elements = test_element.get('elements')
            stations = ghcn_daily.get_stations(elements=elements,
                as_dataframe=True)
            _check_stations_dataframe(stations,
                test_element.get('includes'),
                test_element.get('excludes'))


def _check_stations_dataframe(stations, includes, excludes):
    for include in includes:
        assert include in stations['id']
    for exclude in excludes:
        assert not exclude in stations['id']



"""
   ulmo.util
   ~~~~~~~~~~

   Collection of useful functions for common use cases
"""
from contextlib import contextmanager
import os
import warnings

import tables

from . import misc as util_misc


def get_default_h5file_path(dataset):
    default_dir = util_misc.get_ulmo_dir()
    return os.path.join(default_dir, 'ulmo_%s.h5' % dataset)


def get_or_create_group(h5file, path, title, *args, **kwargs):
    return _get_or_create_node('createGroup', h5file, path, title, *args,
            **kwargs)


def get_or_create_table(h5file, path, table_definition, title):
    return _get_or_create_node('createTable', h5file, path, table_definition,
            title)


@contextmanager
def open_h5file(path, mode, filters=None):
    """returns an open h5file, creating a new one if it doesn't
    already exist
    """
    if filters is None:
        filters = _best_available_filters(['blosc', 'zlib'])

    # create file if it doesn't exist
    util_misc.mkdir_if_doesnt_exist(os.path.dirname(path))
    if not os.path.exists(path):
        new_file = tables.openFile(path, mode='w', title='ulmo data', filters=filters)
        new_file.close()

    open_file = tables.openFile(path, mode=mode)
    yield open_file
    open_file.close()


def update_or_append_sortable(table, update_values, sortby):
    """updates table with dict representations of rows, appending new rows if
    need be; sortby should be a completly sortable column (with a CSIndex)
    """
    value_row = table.row
    update_values.sort(key=lambda v: v[sortby])
    table_iterator = table.itersorted(sortby)
    try:
        current_row = next(table_iterator)
    except StopIteration:
        current_row = None

    for i, update_value in enumerate(update_values):
        if not current_row or update_value[sortby] < current_row[sortby]:
            update_value['__flag_for_append'] = True

        elif current_row:
            # advance the table iterator until you are >= update_value
            while current_row and current_row[sortby] < update_value[sortby]:
                try:
                    current_row = next(table_iterator)
                except StopIteration:
                    current_row = None

            # if we match, then update
            if current_row and current_row[sortby] == update_value[sortby]:
                _update_row_with_dict(current_row, update_value)
                current_row.update()

            # else flag for append
            else:
                update_value['__flag_for_append'] = True

    for update_value in update_values:
        if '__flag_for_append' in update_value:
            del update_value['__flag_for_append']
            _update_row_with_dict(value_row, update_value)
            value_row.append()
    table.flush()


def _best_available_filters(possible_compressions):
    """returns the best available filters to use"""
    for possible_compression in possible_compressions:
        try:
            return tables.Filters(complevel=1, complib=possible_compression)
        except tables.FiltersWarning:
            pass

    return tables.Filters()


def _update_row_with_dict(row, dict):
    """sets the values of row to be the values found in dict"""
    for k, v in dict.items():
        row.__setitem__(k, v)


def _get_or_create_node(method_name, h5file, path, *args, **kwargs):
    try:
        node = h5file.getNode(path)
    except tables.exceptions.NoSuchNodeError:
        where, name = path.rsplit('/', 1)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            create_method = getattr(h5file, method_name)
            node = create_method(where, name, *args, **kwargs)
    return node



"""
   ulmo.util.raster
   ~~~~~~~~~~~~~~~~

   Collection of useful functions for raster manipulation
"""
from __future__ import print_function
from past.builtins import basestring

import contextlib
import hashlib
from .misc import download_if_new, mkdir_if_doesnt_exist
import os
import zipfile


def mosaic_and_clip(raster_tiles, xmin, ymin, xmax, ymax, output_path):
    from pyproj import Proj
    import rasterio
    import subprocess

    print('Mosaic and clip to bounding box extents')
    output_vrt = os.path.splitext(output_path)[0] + '.vrt'
    print(subprocess.check_output(['gdalbuildvrt', '-overwrite', output_vrt] + raster_tiles))
    # check crs
    with rasterio.drivers():
        with rasterio.open(output_vrt) as src:
            p = Proj(src.crs)

    if not p.is_latlong():
        [xmax, xmin],[ymax, ymin] = p([xmax, xmin],[ymax,ymin])

    print(subprocess.check_output(['gdalwarp', '-overwrite', '-te', repr(xmin), repr(ymin), repr(xmax), repr(ymax), output_vrt, output_path]))
    print('Output raster saved at %s', output_path)


def download_tiles(path, tile_urls, tile_fmt, check_modified=False):
    raster_tiles = []

    if isinstance(tile_urls, basestring):
        tile_urls = [tile_urls]

    for i, url in enumerate(tile_urls):
        filename = os.path.split(url)[-1]
        print('... downloading tile %s of %s from %s' % (i+1, len(tile_urls), url))
        mkdir_if_doesnt_exist(path)
        mkdir_if_doesnt_exist(os.path.join(path,'zip'))
        tile_path = os.path.join(path, filename)
        if tile_fmt=='':
            download_if_new(url, tile_path, check_modified=check_modified)
        else:
            zip_path = os.path.join(path, 'zip', filename)
            download_if_new(url, zip_path, check_modified=check_modified)
            print('... ... zipfile saved at %s' % zip_path)
            tile_path = extract_from_zip(zip_path, tile_path, tile_fmt)

        raster_tiles.append(tile_path)
    return raster_tiles


def extract_from_zip(zip_path, tile_path, tile_fmt):
    tile_path = os.path.splitext(tile_path)[0] + tile_fmt
    with zipfile.ZipFile(zip_path) as z:
        fname = [x for x in z.namelist() if tile_fmt in x[-4:]][0]
        with open(tile_path, 'wb') as f:
            f.write(z.read(fname))
            print('... ... %s format raster saved at %s' % (tile_fmt, tile_path))

    return tile_path


def generate_raster_uid(layer, xmin, ymin, xmax, ymax):
    return hashlib.md5(','.join([layer, repr(xmin), repr(ymin), repr(xmax), repr(ymax)])).hexdigest()



# coding: utf-8

# # Using ulmo to display Daymet data for individual locations

# 
# Daymet is a a daily gridded weather dataset distributed through the ORNL DAAC https://daymet.ornl.gov/
# 
# As part of their tools for obtaining data they have created a webservice that allows one to download a timeseries of temperature and precipitation for any location in the US.
# 

# In[1]:

from ulmo.nasa import daymet


# In[2]:

ornl_lat, ornl_long = 35.9313167, -84.3104124
df = daymet.get_daymet_singlepixel(longitude=ornl_long, latitude=ornl_lat, 
                                   years=[2012,2013])


# ### Which gives us a dataframe with daily weather data for the Oak Ridge National Lab

# In[3]:

df.head()


# ### Which we can visualize using matplotlib and seaborn

# In[4]:

import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt


# In[5]:

get_ipython().magic('matplotlib inline')
fig, (ax1, ax2) = plt.subplots(2, figsize=(18, 10), sharex=True)
rolling15day = df.rolling(center=False,window=15).mean()
ax1.fill_between(rolling15day.index, rolling15day.tmin, rolling15day.tmax, alpha=0.5, lw=0)
ax1.plot(df.index, df[['tmax', 'tmin']].mean(axis=1), lw=2, alpha=0.5)

ax1.set_title('Daymet temp at ORNL', fontsize=20)
ax1.set_ylabel(u'Temp. (°C)', fontsize=20)

monthlysum = df.resample("M").sum()
ax2.bar(monthlysum.index, monthlysum.prcp, width=20,)
ax2.set_title('Daymet precip at ORNL', fontsize=20)
ax2.set_ylabel(u'Precip. (mm)', fontsize=20)

fig.tight_layout()


# ### And ask questions such as: "What does the difference in temperature between Denver and Miami look like?"

# In[6]:

denver_loc = (-104.9903, 39.7392)
miami_loc = (-80.2089, 25.7753)

denver = daymet.get_daymet_singlepixel(longitude=denver_loc[0], latitude=denver_loc[1], 
                                   years=[2012, 2013, 2014])
miami = daymet.get_daymet_singlepixel(longitude=miami_loc[0], latitude=miami_loc[1], 
                                   years=[2012, 2013, 2014])


# In[9]:

sns.set_context("talk")
fig, ax1 = plt.subplots(1, figsize=(18, 10))
den_15day = denver.rolling(center=False,window=15).mean()
ax1.fill_between(den_15day.index, den_15day.tmin, den_15day.tmax, 
                 alpha=0.4, lw=0, label='Denver', color=sns.xkcd_palette(['faded green'])[0])

ax1.set_title('Denver vs Miami temps (15 day rolling mean)', fontsize=20)

miami_15day = miami.rolling(center=False,window=15).mean()
ax1.fill_between(miami_15day.index, miami_15day.tmin, miami_15day.tmax, 
                 alpha=0.4, lw=0, label='Miami', color=sns.xkcd_palette(['dusty purple'])[0])

ax1.set_ylabel(u'Temp. (°C)', fontsize=20)
fig.tight_layout()
plt.legend(fontsize=20)




"""
    NOAA `GOES Data Collection System`_ 
    Access to data stream transmitted via GOES satellite.

    .. _GOES Data Collection System: https://www.noaasis.noaa.gov/GOES/GOES_DCS/goes_dcs.html
"""

from .core import get_data, decode



from builtins import zip
from past.builtins import basestring
import contextlib
import os
import os.path
import re
import shutil
import tempfile

import httpretty as httpretty
import mock


def get_test_file_path(file_path):
    """translates a file path to be relative to the test files directory"""
    return os.path.join(os.path.dirname(__file__), 'files', file_path)


@contextlib.contextmanager
def mocked_suds_client(waterml_version, mocked_service_calls, force=False):
    """mocks the suds library to return a given file's content"""
    # if environment variable is set, then mock the tests otherwise just grab files
    # over the network. Example:
    #    env ULMO_MOCK_TESTS=1 py.test
    if not os.environ.get('ULMO_MOCK_TESTS', False) and not force:
        yield

    else:
        tns_str = 'http://www.cuahsi.org/his/%s/ws/' % waterml_version
        with _open_multiple(list(mocked_service_calls.values())) as open_files:
            client = mock.MagicMock()
            client.wsdl.tns = ('tns', tns_str)

            for service_call, filename in mocked_service_calls.items():
                open_file = open_files[filename]

                def _func(*args, **kwargs):
                    return open_file.read()
                setattr(client.service, service_call, _func)

            with mock.patch('suds.client.Client', return_value=client):
                yield


@contextlib.contextmanager
def mocked_urls(url_files, methods=None, force=False):
    """mocks the underlying python sockets library to return a given file's
    content. Note: that this function checks for an environment variable named
    ULMO_DONT_MOCK_TESTS; if that environment variable is set then urls will
    not be mocked and the HTTP requests will go over the network. For example,
    this could be used to to run the whole test suite without mocking files:
    #    env ulmo_dont_mock_tests=1 py.test

    Parameters
    ----------
    url_files : str or dict
        Files to be mocked. It can either be a string representation of a
        filepath to a file whose contents will be used as a response to all
        HTTP requests. Or it can be a dict where the keys are regular
        expression strings and the values are filepaths - the regex keys will
        be used to match urls if they match, the file path will be used.

    methods : iterable of str or None
        HTTP methods that will be mocked. If set to None (default) then the
        default methods are GET, POST and HEAD.
    """
    # if environment variable is set, then mock the tests otherwise just grab files
    # over the network. Example:
    #    env ULMO_MOCK_TESTS=1 py.test
    if not os.environ.get('ULMO_MOCK_TESTS', False) and not force:
        yield

    else:
        if isinstance(url_files, basestring):
            url_files = {'.*': url_files}

        httpretty.enable()
        for url_match, url_file in url_files.items():
            if not isinstance(url_match, basestring) and len(url_match) == 2:
                url_match, methods = url_match

            if not os.path.isabs(url_file):
                url_file = get_test_file_path(url_file)

            callback = _build_request_callback(url_file)
            url_re = re.compile(url_match)

            if methods is None:
                methods = ['GET', 'POST', 'HEAD']

            for method in methods:
                request_class = getattr(httpretty, method)
                httpretty.register_uri(request_class, url_re, match_querystring=True, body=callback)
        yield
        httpretty.disable()
        httpretty.reset()


@contextlib.contextmanager
def temp_dir():
    temp_dir = tempfile.mkdtemp()
    yield temp_dir
    shutil.rmtree(temp_dir)


def use_test_files():
    """Returns true if tests should be run using test files, false otherwise."""
    return os.environ.get('ULMO_DONT_MOCK_TESTS', True)


def _build_request_callback(response_file):
    def request_callback(method, uri, headers):
        if method == 'HEAD':
            response_text = ''
        else:
            with open(response_file, 'rb') as f:
                response_text = f.read()
        return [200, headers, response_text]
    return request_callback


@contextlib.contextmanager
def _open_multiple(paths, handlers=None, new_paths=None):
    """context manager that yields a dict containing file paths as keys mapped
    to their corresponding open file handles
    """
    # note: recursion is the pattern here
    if not handlers and not new_paths:
        with _open_multiple([], [], paths) as return_dict:
            yield return_dict
    elif len(new_paths) == 0:
        yield dict(list(zip(paths, handlers)))
    else:
        next_path = new_paths.pop(0)
        test_path = get_test_file_path(next_path)
        with open(test_path, 'rb') as f:
            paths.append(next_path)
            handlers.append(f)
            with _open_multiple(paths, handlers, new_paths) as return_dict:
                yield return_dict



"""
    ulmo.usgs.ned.core
    ~~~~~~~~~~~~~~~~~~~~~

    This module provides access to the `National Elevation Dataset` provided by the `United States Geological
    Survey`_ `National Map`_ system.

    .. _National Elevation Dataset: http://ned.usgs.gov/
    .. _United States Geological Survey: http://www.usgs.gov/
    .. _National Map: http://nationalmap.gov

"""
from __future__ import print_function
from past.builtins import basestring

from geojson import Feature, FeatureCollection, Polygon
import logging
import os
import requests
from ulmo import util


# NED ftp url.
NED_FTP_URL = 'ftp://rockyftp.cr.usgs.gov/vdelivery/Datasets/Staged/NED/<layer>/IMG/'

# default file path (appended to default ulmo path)
DEFAULT_FILE_PATH = 'usgs/ned/'

layer_dict = {
        'Alaska 2 arc-second': '4f70aaece4b058caae3f8de9',
        '1 arc-second': '4f70aa71e4b058caae3f8de1',
        '1/3 arc-second': '4f70aa9fe4b058caae3f8de5',
        '1/9 arc-second': '4f70aac4e4b058caae3f8de7',
    }

# configure logging
LOG_FORMAT = '%(message)s'
logging.basicConfig(format=LOG_FORMAT)
log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


def get_available_layers():
    """return list of available data layers
    """
    return list(layer_dict.keys())


def get_raster_availability(layer, bbox=None):
    """retrieve metadata for raster tiles that cover the given bounding box
    for the specified data layer.

    Parameters
    ----------
    layer : str
        dataset layer name. (see get_available_layers for list)
    bbox : (sequence of float|str)
        bounding box of in geographic coordinates of area to download tiles
        in the format (min longitude, min latitude, max longitude, max latitude)

    Returns
    -------
    metadata : geojson FeatureCollection
        returns metadata including download urls as a FeatureCollection
    """

    base_url = 'https://www.sciencebase.gov/catalog/items'
    params = [
        ('parentId', layer_dict[layer]),
        ('filter', 'tags=IMG'),
        ('max', 1000),
        ('fields', 'webLinks,spatial,title'),
        ('format', 'json'),
    ]

    if bbox:
        xmin, ymin, xmax, ymax = [float(n) for n in bbox]
        polygon = 'POLYGON (({}))'.format(','.join([(repr(x) + ' ' + repr(y)) for x,y in [
            (xmin, ymax),
            (xmin, ymin),
            (xmax, ymin),
            (xmax, ymax),
            (xmin, ymax)]]))
        params.append(('filter', 'spatialQuery={{wkt:"{}",relation:"{}"}}'.format(polygon, 'intersects')))

    features = []
    url = base_url
    while url:
        r = requests.get(url, params)
        print('retrieving raster availability from %s' % r.url)
        params = []  # not needed after first request
        content = r.json()
        for item in content['items']:
            feature = Feature(geometry=Polygon(_bbox2poly(item['spatial']['boundingBox'])), id=item['id'],
                        properties={
                            'name': item['title'],
                            'layer': layer,
                            'format': '.img',
                            'download url': [x for x in item['webLinks'] if x['type']=='download'][0]['uri']}
                    )
            features.append(feature)

        if content.get('nextlink'):
            url = content['nextlink']['url']
        else:
            break

    return FeatureCollection(features)


def get_raster(layer, bbox, path=None, update_cache=False,
               check_modified=False, mosaic=False):
    """downloads National Elevation Dataset raster tiles that cover the given bounding box
    for the specified data layer.

    Parameters
    ----------
    layer : str
        dataset layer name. (see get_available_layers for list)
    bbox : (sequence of float|str)
        bounding box of in geographic coordinates of area to download tiles
        in the format (min longitude, min latitude, max longitude, max latitude)
    path : ``None`` or path
        if ``None`` default path will be used
    update_cache: ``True`` or ``False`` (default)
        if ``False`` and output file already exists use it.
    check_modified: ``True`` or ``False`` (default)
        if tile exists in path, check if newer file exists online and download if available.
    mosaic: ``True`` or ``False`` (default)
        if ``True``, mosaic and clip downloaded tiles to the extents of the bbox provided. Requires
        rasterio package and GDAL.

    Returns
    -------
    raster_tiles : geojson FeatureCollection
        metadata as a FeatureCollection. local url of downloaded data is in feature['properties']['file']
    """
    _check_layer(layer)

    raster_tiles = _download_tiles(get_raster_availability(layer, bbox), path=path,
        check_modified=check_modified)

    if mosaic:
        if path is None:
            path = os.path.join(util.get_ulmo_dir(), DEFAULT_FILE_PATH)

        util.mkdir_if_doesnt_exist(os.path.join(path, 'by_boundingbox'))
        xmin, ymin, xmax, ymax = [float(n) for n in bbox]
        uid = util.generate_raster_uid(layer, xmin, ymin, xmax, ymax)
        output_path = os.path.join(path, 'by_boundingbox', uid + '.tif')

        if os.path.isfile(output_path) and not update_cache:
            return output_path

        raster_files = [tile['properties']['file'] for tile in raster_tiles['features']]
        util.mosaic_and_clip(raster_files, xmin, ymin, xmax, ymax, output_path)
        return [output_path]

    return raster_tiles

def _check_layer(layer):
    """
    make sure the passed layer name is one of the handled options
    """

    if not layer in get_available_layers():
        err_msg = "The specified layer parameter ({})".format(layer)
        err_msg += "\nis not in the available options:"
        err_msg += "\n\t".join(get_available_layers())
        raise ValueError(err_msg)


def _download_features(feature_ids, path=None, check_modified=False,):
    if path is None:
        path = os.path.join(util.get_ulmo_dir(), DEFAULT_FILE_PATH)

    if isinstance(feature_ids, basestring):
        feature_ids = [feature_ids]

    tiles =[]
    tile_fmt = '.img'
    for feature_id in feature_ids:
        url = SCIENCEBASE_ITEM_URL % feature_id
        metadata = requests.get(url).json()
        layer = [a for a in list(layer_dict.keys()) if a in metadata['title']][0]
        layer_path = os.path.join(path, layer_dict[layer])
        tile_urls = [link['uri'] for link in metadata['webLinks'] if link['type']=='download']
        tiles.append({'feature_id': feature_id,
                      'tiles': util.download_tiles(layer_path, tile_urls, tile_fmt, check_modified),
                      })

    return tiles


def _bbox2poly(bbox):
    xmin = bbox['minX']
    xmax = bbox['maxX']
    ymin = bbox['minY']
    ymax = bbox['maxY']

    return [[(xmin,ymin), (xmin,ymax), (xmax, ymax), (xmax, ymin), (xmin, ymin)]]


def _download_tiles(tiles, path=None, check_modified=False):

    if path is None:
        path = os.path.join(util.get_ulmo_dir(), DEFAULT_FILE_PATH)

    for tile in tiles['features']:

        metadata = tile['properties']
        layer_path = os.path.join(path, layer_dict[metadata['layer']])
        tile['properties']['file'] = util.download_tiles(layer_path, metadata['download url'], metadata['format'], check_modified)[0]

    return tiles



"""
    `National Climatic Data Center`_ `Global Summary of the Day`_ dataset

    .. _National Climatic Data Center: http://www.ncdc.noaa.gov
    .. _Global Summary of the Day: http://www.ncdc.noaa.gov/oa/gsod.html
"""
from .core import (get_data, get_stations)



import pytest

import ulmo
import test_util


def test_get_stations():
    stations_file = 'usace/swtwc/shefids.html'
    with test_util.mocked_urls(stations_file):
        stations = ulmo.usace.swtwc.get_stations()

    test_stations = [
        {'code': u'DSNT2', 'description': u'Lake Texoma, Denison Dam'},
        {'code': u'MYST2', 'description': u'Pat Mayse Lake'},
    ]

    for test_station in test_stations:
        assert stations[test_station['code']] == test_station
    assert 700 <= len(stations) <= 800


def test_get_station_data():
    test_station_data = [
        ('DSNT2', '2017-07-25', {
            'code': 'DSNT2',
            'description': 'Lake Texoma, Denison Dam',
            'station_type': 'Reservoir',
            'timezone': 'US/Central',
            'values': {
                '2017-07-25 01:00:00': {
                    'AIR-TEMP': 80.10,
                    'BAT-LOAD': 11.37,
                    'ELEVATION': 617.73,
                    'INFLOW': 564,
                    'PRECIP': 0.0,
                    'PRECIP(A)': 0.0,
                    'REL-HUMID': 81.98,
                    'RELEASE': 19,
                    'SOLAR-RAD': 0.0,
                    'STORAGE': 2572324.0,
                    'VOLTAGE': 11.77,
                    'WIND-DIR': 135.90,
                    'WIND-SPEED': 1.42,
                    'TW-ELEV': 501.87
                }
            },
        }),
    ]

    for code, date, test_data in test_station_data:
        url_date = date.replace('-', '')
        filename = '%s.%s.html' % (code, url_date)
        data_file = 'usace/swtwc/' + filename
        with test_util.mocked_urls(data_file):
            station_data = ulmo.usace.swtwc.get_station_data(code, date)

        for key, value in test_data.items():
            if key == 'values':
                _compare_values(test_data['values'], station_data['values'])
            else:
                assert station_data[key] == test_data[key]


def test_get_station_data_current():
    # can't easily test current since it is a moving target changes, but mostly
    # just make sure it parses correctl: current will have '---' values where
    # previous days do not
    data_file = 'usace/swtwc/DSNT2.current.html'
    with test_util.mocked_urls(data_file):
        station_data = ulmo.usace.swtwc.get_station_data('DSNT2')
    assert len(station_data.get('values')) > 0


def test_get_station_data_out_of_range():
    # can't easily test current since it is a moving target changes, but mostly
    # just make sure it parses correctl: current will have '---' values where
    # previous days do not
    data_file = 'usace/swtwc/empty.html'
    with test_util.mocked_urls(data_file):
        with pytest.raises(ValueError):
            station_data = ulmo.usace.swtwc.get_station_data('MYST2', '1945-01-01')


def _compare_values(test_values, station_values):
    for key, test_value in test_values.items():
        assert station_values[key] == test_value



from .misc import (
        camel_to_underscore,
        convert_date,
        convert_datetime,
        dict_from_dataframe,
        dir_list,
        download_if_new,
        get_ulmo_dir,
        mkdir_if_doesnt_exist,
        module_with_dependency_errors,
        module_with_deprecation_warnings,
        open_file_for_url,
        parse_fwf,
        raise_dependency_error,
        save_pretty_printed_xml,
        to_bytes,
    )

from .raster import (
        extract_from_zip,
        mosaic_and_clip,
        download_tiles,
        generate_raster_uid,
    )

try:
    from .pytables import (
            get_default_h5file_path,
            get_or_create_group,
            get_or_create_table,
            open_h5file,
            update_or_append_sortable,
        )
except ImportError:
    get_default_h5file_path = raise_dependency_error
    get_or_create_group = raise_dependency_error
    get_or_create_table = raise_dependency_error
    open_h5file = raise_dependency_error
    update_or_append_sortable = raise_dependency_error



"""
    ulmo.ncdc.gsod.core
    ~~~~~~~~~~~~~~~~~~~

    This module provides direct access to `National Climatic Data Center`_
    `Global Summary of the Day`_ dataset.

    .. _National Climatic Data Center: http://www.ncdc.noaa.gov
    .. _Global Summary of the Day: http://www.ncdc.noaa.gov/oa/gsod.html
"""
from builtins import str
from builtins import range
from past.builtins import basestring
from contextlib import contextmanager
import csv
import datetime
import gzip
import itertools
import os
import tarfile

import numpy as np

from ulmo import util

NCDC_GSOD_DIR = os.path.join(util.get_ulmo_dir(), 'ncdc/gsod')
NCDC_GSOD_STATIONS_FILE = os.path.join(NCDC_GSOD_DIR, 'isd-history.csv')
NCDC_GSOD_START_DATE = datetime.date(1929, 1, 1)

def get_parameters():
    """
    retrieve a list of parameter codes available.
    Reference for GSOD parameters : https://www1.ncdc.noaa.gov/pub/data/gsod/readme.txt

    Parameters
    ----------
        None

    Returns
    -------
        dictionary of variables with parameter codes as keys
        and GSOD codes as values.
    """

    VARIABLES = {'mean_temp' : 'TEMP',
                'mean_temp_count' : 'TEMP',
                'dew_point' : 'DEWP',
                'dew_point_count' : 'DEWP',
                'sea_level_pressure' : 'SLP     ',
                'sea_level_pressure_count' : 'SLP     ',
                'station_pressure' : 'STP',
                'station_pressure_count' : 'STP',
                'visibility' : 'VISIB',
                'visibility_count' : 'VISIB',
                'mean_wind_speed' : 'WDSP',
                'mean_wind_speed_count' : 'WDSP',
                'max_wind_speed' : 'MXSPD',
                'max_gust' : 'GUST',
                'max_temp' : 'MAX',
                'max_temp_flag' : 'MAX',
                'min_temp' : 'MIN',
                'min_temp_flag' : 'MIN',
                'precip' : 'PRCP',
                'precip_flag' : 'PRCP',
                'snow_depth' : 'SNDP',
                'FRSHTT' : 'FRSHTT'}
    return VARIABLES

def get_data(station_codes, start=None, end=None, parameters=None):
    """Retrieves data for a set of stations.


    Parameters
    ----------
    station_codes : str or list
        Single station code or iterable of station codes to retrieve data for.
    start : ``None`` or date (see :ref:`dates-and-times`)
        If specified, data are limited to values after this date.
    end : ``None`` or date (see :ref:`dates-and-times`)
        If specified, data are limited to values before this date.
    parameters : ``None``, str or list
        If specified, data are limited to this set of parameter codes.


    Returns
    -------
    data_dict : dict
        Dict with station codes keyed to lists of value dicts.
    """
    if start:
        start_date = util.convert_date(start)
    else:
        start_date = NCDC_GSOD_START_DATE
    if end:
        end_date = util.convert_date(end)
    else:
        end_date = datetime.date.today()
    if isinstance(parameters, basestring):
        parameters = [parameters]
    if parameters and not 'date' in parameters:
        # add date to list of parameters if it's not there already
        parameters.insert(0, 'date')
    if isinstance(station_codes, basestring):
        station_codes = [station_codes]

    # note: opening tar files and parsing the headers and such is a relatively
    # lengthy operation so you don't want to do it too often, hence try to
    # grab all stations at the same time per tarfile
    data_dict = dict([(station_code, None) for station_code in station_codes])

    for year in range(start_date.year, end_date.year + 1):
        tar_path = _get_gsod_file(year)
        with tarfile.open(tar_path, 'r:') as gsod_tar:
            stations_in_file = [
                name.split('./')[-1].rsplit('-', 1)[0]
                for name in gsod_tar.getnames() if len(name) > 1]
            if station_codes:
                stations = list(set(station_codes) & set(stations_in_file))
            else:
                stations = stations_in_file
            for station in stations:
                year_data = _read_gsod_file(gsod_tar, station, year)
                if parameters:
                    year_data = _subset_record_array(year_data, parameters)
                if not year_data is None:
                    # apply date ranges if they exist
                    if start_date or end_date:
                        mask = np.ones(len(year_data), dtype=bool)
                        if start_date:
                            mask = mask & (year_data['date'] >= start_date)
                        if end_date:
                            mask = mask & (year_data['date'] <= end_date)
                        year_data = year_data[mask]

                    if not data_dict[station] is None:
                        # XXX: this could be more efficient for large numbers
                        # of years with a list comprehension or generator
                        data_dict[station] = np.append(data_dict[station], year_data)
                    else:
                        data_dict[station] = year_data
    for station, data_array in data_dict.items():
        if not data_dict[station] is None:
            data_dict[station] = _record_array_to_value_dicts(data_array)
    return data_dict


def get_stations(country=None, state=None, start=None, end=None, update=True):
    """Retrieve information on the set of available stations.


    Parameters
    ----------
    country : {``None``, str, or iterable}
        If specified, results will be limited to stations with matching country
        codes.
    state : {``None``, str, or iterable}
        If specified, results will be limited to stations with matching state
        codes.
    start : ``None`` or date (see :ref:`dates-and-times`)
        If specified, results will be limited to stations which have data after
        this start date.
    end : ``None`` or date (see :ref:`dates-and-times`)
        If specified, results will be limited to stations which have data before
        this end date.
    update : bool
        If ``True`` (default), check for a newer copy of the stations file and
        download if it is newer the previously downloaded copy. If ``False``,
        then a new stations file will only be downloaded if a previously
        downloaded file cannot be found.


    Returns
    -------
    stations_dict : dict
        A dict with USAF-WBAN codes keyed to station information dicts.
    """
    if start:
        start_date = util.convert_date(start)
    else:
        start_date = None
    if end:
        end_date = util.convert_date(end)
    else:
        end_date = None

    if isinstance(country, basestring):
        country = [country]
    if isinstance(state, basestring):
        state = [state]

    stations_url = 'http://www1.ncdc.noaa.gov/pub/data/noaa/isd-history.csv'
    with util.open_file_for_url(stations_url, NCDC_GSOD_STATIONS_FILE) as f:
        reader = csv.DictReader(f)

        if country is None and state is None and start is None and end is None:
            rows = reader
        else:
            if start_date is None:
                start_str = None
            else:
                start_str = start_date.strftime('%Y%m%d')
            if end_date is None:
                end_str = None
            else:
                end_str = end_date.strftime('%Y%m%d')
            rows = [
                row for row in reader
                if _passes_row_filter(row, country=country, state=state,
                    start_str=start_str, end_str=end_str)
            ]

        stations = dict([
            (_station_code(row), _process_station(row))
            for row in rows
        ])
    return stations


def _convert_date_string(date_string):
    if date_string == '':
        return None

    if isinstance(date_string, bytes):
        date_string = date_string.decode('utf-8')

    return datetime.datetime.strptime(date_string, '%Y%m%d').date()


def _get_gsod_file(year):
    url = 'http://www1.ncdc.noaa.gov/pub/data/gsod/%s/gsod_%s.tar' % (year, year)
    filename = url.split('/')[-1]
    path = os.path.join(NCDC_GSOD_DIR, filename)
    util.download_if_new(url, path, check_modified=True)
    return path


def _passes_row_filter(row, country=None, state=None, start_str=None,
        end_str=None):
    if not country is None and row['CTRY'] not in country:
        return False
    if not state is None and row['STATE'] not in state:
        return False
    if not start_str is None and row['END'] != '' and row['END'] <= start_str:
        return False
    if not end_str is None and row['BEGIN'] != '' and end_str <= row['BEGIN']:
        return False
    return True


def _process_station(station_row):
    """converts a csv row to a more human-friendly version"""
    station_dict = {
        'begin': _convert_date_string(station_row['BEGIN']),
        'icao': station_row['ICAO'],
        'country': station_row['CTRY'],
        'elevation': round(float(station_row['ELEV(M)']), 1) \
                if station_row['ELEV(M)'] not in ('', '-99999') else None,
        'end': _convert_date_string(station_row['END']),
        'latitude': round(float(station_row['LAT']), 3) \
                if station_row['LAT'] not in ('', '-99999') else None,
        'longitude': round(float(station_row['LON']), 3) \
                if station_row['LON'] not in ('', '-999999') else None,
        'name': station_row['STATION NAME'],
        'state': station_row['STATE'],
        'USAF': station_row['USAF'],
        'WBAN': station_row['WBAN'],
    }
    return station_dict


def _read_gsod_file(gsod_tar, station, year):
    tar_station_filename = station + '-' + str(year) + '.op.gz'
    try:
        gsod_tar.getmember('./' + tar_station_filename)
    except KeyError:
        return None

    ncdc_temp_dir = os.path.join(NCDC_GSOD_DIR, 'temp')
    util.mkdir_if_doesnt_exist(ncdc_temp_dir)
    temp_path = os.path.join(ncdc_temp_dir, tar_station_filename)

    gsod_tar.extract('./' + tar_station_filename, ncdc_temp_dir)
    with gzip.open(temp_path, 'rb') as gunzip_f:
        columns = [
            # name, length, # of spaces separating previous column, dtype
            ('USAF', 6, 0, 'U6'),
            ('WBAN', 5, 1, 'U5'),
            ('date', 8, 2, object),
            ('mean_temp', 6, 2, float),
            ('mean_temp_count', 2, 1, int),
            ('dew_point', 6, 2, float),
            ('dew_point_count', 2, 1, int),
            ('sea_level_pressure', 6, 2, float),
            ('sea_level_pressure_count', 2, 1, int),
            ('station_pressure', 6, 2, float),
            ('station_pressure_count', 2, 1, int),
            ('visibility', 5, 2, float),
            ('visibility_count', 2, 1, int),
            ('mean_wind_speed', 5, 2, float),
            ('mean_wind_speed_count', 2, 1, int),
            ('max_wind_speed', 5, 2, float),
            ('max_gust', 5, 2, float),
            ('max_temp', 6, 2, float),
            ('max_temp_flag', 1, 0, 'U1'),
            ('min_temp', 6, 1, float),
            ('min_temp_flag', 1, 0, 'U1'),
            ('precip', 5, 1, float),
            ('precip_flag', 1, 0, 'U1'),
            ('snow_depth', 5, 1, float),
            ('FRSHTT', 6, 2, 'U6'),
        ]

        dtype = np.dtype([
            (column[0], column[3])
            for column in columns])

        # note: ignore initial 0
        delimiter = itertools.chain(*[column[1:3][::-1] for column in columns])
        usecols = list(range(1, len(columns) * 2, 2))

        data = np.genfromtxt(gunzip_f, skip_header=1, delimiter=delimiter,
                usecols=usecols, dtype=dtype, converters={5: _convert_date_string})
    os.remove(temp_path)

    # somehow we can end up with single-element arrays that are 0-dimensional??
    # (occurs on tyler's machine but is hard to reproduce)
    if data.ndim == 0:
        data = data.flatten()

    return data


def _record_array_to_value_dicts(record_array):
    names = record_array.dtype.names
    value_dicts = [
        dict([(name, value[name_index])
                for name_index, name in enumerate(names)])
        for value in record_array]
    return value_dicts


def _station_code(station):
    """returns station code from a station dict"""
    return '-'.join([station['USAF'], station['WBAN']])


def _subset_record_array(record_array, parameters):
    dtype = np.dtype([
        (parameter, record_array[parameter].dtype)
        for parameter in parameters
    ])
    return np.array([
        tuple(i) for i in np.array([
            record_array[parameter] for parameter in parameters
        ]).T.tolist()
    ], dtype=dtype)



from future import standard_library
standard_library.install_aliases()
from builtins import zip
from builtins import str
from past.builtins import basestring
from builtins import object
from contextlib import contextmanager
import datetime
import email.utils
import ftplib
import functools
import os
import re
import urllib.parse
import warnings

import appdirs
from lxml import etree
import numpy as np
import pandas
import requests


# pre-compiled regexes for underscore conversion
first_cap_re = re.compile('(.)([A-Z][a-z]+)')
all_cap_re = re.compile('([a-z0-9])([A-Z])')


class DependencyError(Exception):
    pass


def camel_to_underscore(s):
    """converts camelCase to underscore, originally from
    http://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-camel-case
    """
    first_sub = first_cap_re.sub(r'\1_\2', s)
    return all_cap_re.sub(r'\1_\2', first_sub).lower()


def convert_date(date):
    """returns a datetime.date object from either a string representation or
    date-like object (datetime.date, datetime.datetime, or pandas.Timestamp)
    """
    return pandas.Timestamp(date).date()


def convert_datetime(datetime):
    """returns a datetime.date object from either a string representation or
    datetime-like object (datetime.date, datetime.datetime, or pandas.Timestamp)
    """
    return pandas.Timestamp(datetime).to_pydatetime()


def dir_list(url):
    """given a path to a ftp directory, returns a list of files in that
    directory
    """
    parsed = urllib.parse.urlparse(url)
    ftp = ftplib.FTP(parsed.netloc, "anonymous")
    ftp.cwd(parsed.path)
    return ftp.nlst()


def dict_from_dataframe(dataframe):
    if isinstance(dataframe.index, pandas.PeriodIndex):
        dataframe.index = dataframe.index.to_timestamp().astype('str')
    if isinstance(dataframe.index, pandas.DatetimeIndex):
        dataframe.index = [str(i) for i in dataframe.index]

    # convert np.nan objects to None objects; prior to pandas 0.13.0 this could
    # be done in a vectorized way, but as of 0.13, assigning None into a
    # dataframe, it gets converted to a nan object so this has to be done
    # rather inefficiently in a post-processing step
    if pandas.__version__ < '0.13.0':
        for column_name in dataframe.columns:
            dataframe[column_name][pandas.isnull(dataframe[column_name])] = None
        df_dict = dataframe.T.to_dict()
    else:
        df_dict = dataframe.where((pandas.notnull(dataframe)), None).to_dict(orient='index')

    return df_dict


def download_if_new(url, path, check_modified=True):
    """downloads the file located at `url` to `path`, if check_modified is True
    it will only download if the url's last-modified header has a more recent
    date than the filesystem's last modified date for the file
    """
    parsed = urllib.parse.urlparse(url)

    if os.path.exists(path) and not check_modified:
        return

    if parsed.scheme.startswith('ftp'):
        _ftp_download_if_new(url, path, check_modified)
    elif parsed.scheme.startswith('http'):
        _http_download_if_new(url, path, check_modified)
    else:
        raise NotImplementedError("only ftp and http urls are currently implemented")


def get_ulmo_dir(sub_dir=None):
    return_dir = appdirs.user_data_dir('ulmo', 'ulmo')
    if sub_dir:
        return_dir = os.path.join(return_dir, sub_dir)
    mkdir_if_doesnt_exist(return_dir)
    return return_dir


def mkdir_if_doesnt_exist(dir_path):
    """makes a directory if it doesn't exist"""
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)


def module_with_dependency_errors(method_names):
    class FakeModule(object):
        pass
    fake_module = FakeModule()

    for method_name in method_names:
        def f(*args, **kwargs):
            raise_dependency_error()
        f.__name__ = method_name
        setattr(fake_module, method_name, f)
    return fake_module


def module_with_deprecation_warnings(functions, warning_message):
    warnings.filterwarnings("always", warning_message, DeprecationWarning)

    class DeprecatedModule(object):
        pass
    deprecated_module = DeprecatedModule()

    def warning_decorator(f):
        @functools.wraps(f)
        def warning_wrapper(*args, **kwargs):
            warnings.warn(warning_message, DeprecationWarning)
            return f(*args, **kwargs)
        return warning_wrapper

    for function in functions:
        wrapped_function = warning_decorator(function)
        setattr(deprecated_module, function.__name__, wrapped_function)
    return deprecated_module


@contextmanager
def open_file_for_url(url, path, check_modified=True, use_file=None, use_bytes=None):
    """Context manager that returns an open file handle for a data file;
    downloading if necessary or otherwise using a previously downloaded file.
    File downloading will be short-circuited if use_file is either a file path
    or an open file-like object (i.e. file handler or StringIO obj), in which
    case the file handler pointing to use_file is returned - if use_file is a
    file handler then the handler won't be closed upon exit.
    """
    leave_open = False

    if use_file is not None:
        if hasattr(use_file, 'read'):
            leave_open = True
            yield use_file
        else:
            open_path = use_file
    else:
        download_if_new(url, path, check_modified)
        open_path = path

    if use_bytes is None:
        open_file = open(open_path, 'r')
    else:
        open_file = open(open_path, 'rb')

    yield open_file

    if not leave_open:
        open_file.close()


def parse_fwf(file_path, columns, na_values=None):
    """Convenience function for parsing fixed width formats. Wraps the pandas
    read_fwf parser but allows all column information to be kept together.
    Columns should be an iterable of lists/tuples with the format (column_name,
    start_value, end_value, converter). Returns a pandas dataframe.
    """
    names, colspecs = list(zip(*[(name, (start, end))
        for name, start, end, converter in columns]))

    converters = dict([
        (name, converter)
        for name, start, end, converter in columns
        if not converter is None
    ])

    return pandas.io.parsers.read_fwf(file_path,
        colspecs=colspecs, header=None, na_values=na_values, names=names,
        converters=converters)


def raise_dependency_error(*args, **kwargs):
    raise DependencyError("Trying to do something that depends on pytables, "
            "but pytables has not been installed.")


def save_pretty_printed_xml(filename, response_buffer):
    """saves a nicely indented version of the xml contained in response_buffer
    to filename; handy for debugging or saving responses for to include in tests"""
    with open(filename, 'w') as f:
        response_buffer.seek(0)
        parsed = etree.parse(response_buffer)
        f.write(etree.tostring(parsed, pretty_print=True))
        response_buffer.seek(0)


def to_bytes(s):
    """convert str to bytes for py 2/3 compat
    """

    if isinstance(s, bytes):
        return s

    return s.encode('utf-8', 'ignore')


def _ftp_download_if_new(url, path, check_modified=True):
    parsed = urllib.parse.urlparse(url)
    ftp = ftplib.FTP(parsed.netloc, "anonymous")
    directory, filename = parsed.path.rsplit('/', 1)
    ftp_last_modified = _ftp_last_modified(ftp, parsed.path)
    ftp_file_size = _ftp_file_size(ftp, parsed.path)

    if not os.path.exists(path) or os.path.getsize(path) != ftp_file_size:
        _ftp_download_file(ftp, parsed.path, path)
    elif check_modified and _path_last_modified(path) < ftp_last_modified:
        _ftp_download_file(ftp, parsed.path, path)


def _ftp_download_file(ftp, ftp_path, local_path):
    with open(local_path, 'wb') as f:
        ftp.retrbinary("RETR " + ftp_path, f.write)


def _ftp_file_size(ftp, file_path):
    ftp.sendcmd('TYPE I')
    return ftp.size(file_path)


def _ftp_last_modified(ftp, file_path):
    timestamp = ftp.sendcmd("MDTM " + file_path).split()[-1]
    return datetime.datetime.strptime(timestamp, '%Y%m%d%H%M%S')


def _http_download_file(url, path):
    request = requests.get(url)
    mkdir_if_doesnt_exist(os.path.dirname(path))
    chunk_size = 64 * 1024
    with open(path, 'wb') as f:
        for content in request.iter_content(chunk_size):
            f.write(content)


def _http_download_if_new(url, path, check_modified):
    head = requests.head(url)
    if not os.path.exists(path) or not _request_file_size_matches(head, path):
        _http_download_file(url, path)
    elif check_modified and _request_is_newer_than_file(head, path):
        _http_download_file(url, path)


def _nans_to_nones(nan_dict):
    """takes a dict and if any values are np.nan then it will replace them with
    None"""
    return dict([
        (k, v) if v is not np.nan else (k, None)
        for k, v in nan_dict.items()
    ])


def _parse_rfc_1123_timestamp(timestamp_str):
    return datetime.datetime(*email.utils.parsedate(timestamp_str)[:6])


def _path_last_modified(path):
    """returns a datetime.datetime object representing the last time the file at
    a given path was last modified
    """
    if not os.path.exists(path):
        return None

    return datetime.datetime.utcfromtimestamp(os.path.getmtime(path))


def _request_file_size_matches(request, path):
    """returns True if request content-length header matches file size"""
    content_length = request.headers.get('content-length')
    if content_length and int(content_length) == os.path.getsize(path):
        return True
    else:
        return False


def _request_is_newer_than_file(request, path):
    """returns true if a request's last-modified header is more recent than a
    file's last modified timestamp
    """
    path_last_modified = _path_last_modified(path)

    if path_last_modified is None:
        return True

    if not request.headers.get('last-modified'):
        warnings.warn('no last-modified date for request: %s, downloading file again' % request.url)
        return True

    request_last_modified = _parse_rfc_1123_timestamp(request.headers.get('last-modified'))
    if request_last_modified > path_last_modified:
        return True
    else:
        return False



import os
import re
import sys

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)


rootpath = os.path.abspath(os.path.dirname(__file__))
with open('README.rst') as f:
    # use README for long description but don't include the link to travis-ci;
    # it seems a bit out of place on pypi since it applies to the development
    # version
    long_description = ''.join([
        line for line in f.readlines()
        ])


# this sets __version__
def version():
    """Get the version number."""
    with open(os.path.join(rootpath, "VERSION.txt")) as v:
        _version = v.read()
    return _version.strip()


__version__ = version()


def parse_requirement(requirement, excludes):
    lib = requirement.strip().split('#')[0]  # strip new lines and comments
    if lib:
        lib_name = re.split(r'[[<>=! ]+', lib)[0]
        if lib_name in excludes:
            return ''
        return lib.strip()
    return ''


def load_requirements(excludes=None):
    if excludes is None:
      excludes = []
    requirements = []
    with open('requirements.txt') as requirements_file:
        line = requirements_file.readline()
        while line:
            line = line.strip()
            while line.endswith('\\'):
                line += requirements_file.readline().strip()
            line = ' '.join(line.split('\\'))
            requirement = parse_requirement(line, excludes)
            if requirement:
                requirements.append(requirement)
            line = requirements_file.readline()
    return requirements


excludes = ['pytest']
required = load_requirements(excludes)


setup(
    name='ulmo',
    version=__version__,
    license='BSD',
    author='Dharhas Pothina',
    author_email='dharhas@gmail.com',
    description='clean, simple and fast access to public hydrology and climatology data',
    long_description=long_description,
    url='https://github.com/ulmo-dev/ulmo/',
    keywords='his ulmo water waterml cuahsi wateroneflow usgs ned',
    packages=find_packages(),
    platforms='any',
    install_requires=required,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    tests_require=[
        'mock',
        'freezegun',
        'pytest',
        'httpretty',
        'html5lib',
    ],
    cmdclass={'test': PyTest},
)



import ulmo
import pandas as pd

import test_util


def test_get_services():
    wsdl = ulmo.cuahsi.his_central.core.HIS_CENTRAL_WSDL_URL.split('?')[0]
    wsdl_file = 'his_central/wsdl.xml'
    service_info_url = 'http://hiscentral.cuahsi.org/webservices/hiscentral.asmx'
    service_info_file = 'his_central/get_services.xml'

    url_files = {
        (wsdl, ('GET',)): wsdl_file,
        (service_info_url, ('POST',)): service_info_file,
    }

    #with test_util.mocked_urls(url_files):
    services = ulmo.cuahsi.his_central.get_services()

    check_services = [
        {'abstract': 'The USGS National Water Information System (NWIS) provides access to millions of sites measuring streamflow, groundwater levels, and water quality. This web service provides methods for retrieving daily values data, such as discharge and water levels, from NWIS. For more information about NWIS data services, see http://waterservices.usgs.gov/',
         'citation': 'U.S. Geological Survey, [YEAR], National Water Information System data, accessed [DATE ACCESSED] via HIS Central (http://hiscentral.cuahsi.org).                                                                                                                                                                                                                    ',
         'email': 'help@cuahsi.org',
         'max_x': -64.69125,
         'max_y': 71.29403,
         'min_x': -176.6633,
         'min_y': -14.3075,
         'network_name': 'NWISDV',
         'organization': 'U.S. Geological Survey (USGS)',
         'organization_website': 'http://www.usgs.gov',
         'phone': '339-221-5400',
         'service_description_url': 'http://hiscentral.cuahsi.org/pub_network.aspx?n=1',
         'service_id': 1,
         'service_status': None,
         'service_url': 'http://hydroportal.cuahsi.org/nwisdv/cuahsi_1_1.asmx?WSDL',
         'site_count': 34841,
         'title': 'NWIS Daily Values',
         'value_count': 387092632,
         'variable_count': 484},
    ]

    assert 89 <= len(services) <= 110

    services = pd.DataFrame(services)
    for check_service in check_services:
        assert check_service['network_name'] in services['network_name'].tolist()
        assert check_service['organization'] in services['organization'].tolist()
        assert check_service['title'] in services['title'].tolist()



"""
    `National Climatic Data Center`_ `Global Historical Climate Network -
    Daily`_ dataset

    .. _National Climatic Data Center: http://www.ncdc.noaa.gov
    .. _Global Historical Climate Network - Daily: http://www.ncdc.noaa.gov/oa/climate/ghcn-daily/
"""
from .core import (get_data, get_stations)



"""
    LCRA Water Quality Data
    ~~~~~~~~~~~~~~~~~~~~~~~
    Access to water quality data in the Colorado River Basin (Texas)
    provided by the `Water Quality`_ web site and web service from
    the `Lower Colorado River Authority`_.

    .. _Lower Colorado River Authority: http://www.lcra.org
    .. _Water Quality: http://waterquality.lcra.org/
"""
from .core import get_sites, get_historical_data, get_recent_data, get_site_info



from ulmo.waterml import common

WATERML_V1_0_NAMESPACE= "{http://www.cuahsi.org/waterML/1.0/}"


def parse_site_infos(content_io):
    """parses site_infos out of a waterml file; content_io should be a file-like object"""
    return common.parse_site_infos(content_io, WATERML_V1_0_NAMESPACE,
            site_info_names=['siteInfo'])


def parse_site_values(content_io, query_isodate=None):
    """parses values out of a waterml file; content_io should be a file-like object"""
    return common.parse_site_values(content_io, WATERML_V1_0_NAMESPACE,
            query_isodate=query_isodate)


def parse_sites(content_io):
    """parses sites out of a waterml file; content_io should be a file-like object"""
    return common.parse_sites(content_io, WATERML_V1_0_NAMESPACE)


def parse_variables(content_io):
    """parses variables out of a waterml file; content_io should be a file-like object"""
    return common.parse_variables(content_io, WATERML_V1_0_NAMESPACE)




import glob
import os

import pandas

import ulmo

import test_util

DATA_DIR = os.path.join(os.path.dirname(__file__), 'files/twc/kbdi')
MOCKED_URLS = dict([
    ('http://twc.tamu.edu/weather_images/summ/' + os.path.basename(path), path)
    for path in glob.glob(os.path.join(DATA_DIR, '*'))
])

test_sets = [
    {
        'start': '2013-04-09',
        'end': '2013-04-09',
        'fips': 48507,
        'values': [{
            'county': 'ZAVALA',
            'date': '2013-04-09',
            'avg': 572,
            'max': 708,
            'min': 423,
        }]
    },
    {
        'start': '2013-04-09',
        'end': '2013-04-09',
        'fips': 48007,
        'values': [{
            'county':  'ARANSAS',
            'date': '2013-04-09',
            'avg': 497,
            'max': 561,
            'min': 439,
        }]
    },
    {
        'start': '2016-10-10',
        'end': '2016-10-10',
        'fips': 48007,
        'values': [{
            'county':  'ARANSAS',
            'date': '2016-10-10',
            'avg': 432,
            'max': 576,
            'min': 179,
        }]
    },
    {
        'start': '2016-10-10',
        'end': '2016-10-10',
        'fips': 48507,
        'values': [{
            'county':  'ZAVALA',
            'date': '2016-10-10',
            'avg': 381,
            'max': 574,
            'min': 127,
        }]
    },
]


def test_get_data_by_county():
    with test_util.temp_dir() as data_dir:
        for test_set in test_sets:
            with test_util.mocked_urls(MOCKED_URLS):
                data = ulmo.twc.kbdi.get_data(
                    county=test_set['fips'],
                    start=test_set['start'],
                    end=test_set['end'],
                    data_dir=data_dir,
                )
            assert len(data) == 1
            assert test_set['fips'] in data


def test_get_data():
    with test_util.temp_dir() as data_dir:
        for test_set in test_sets:
            with test_util.mocked_urls(MOCKED_URLS):
                data = ulmo.twc.kbdi.get_data(
                    start=test_set['start'],
                    end=test_set['end'],
                    data_dir=data_dir,
                )

            values = data.get(test_set['fips'], {})
            test_values = test_set['values']

            for test_value in test_values:
                assert test_value in values


def test_get_data_as_dataframe():
    with test_util.temp_dir() as data_dir:
        with test_util.mocked_urls(MOCKED_URLS):
            data = ulmo.twc.kbdi.get_data(
                start='2013-04-09',
                end='2013-04-09',
                as_dataframe=True,
                data_dir=data_dir,
            )

            assert isinstance(data, pandas.DataFrame)


def test_data_dir_used():
    with test_util.temp_dir() as data_dir:
        with test_util.mocked_urls(MOCKED_URLS):
            ulmo.twc.kbdi.get_data(
                start='2013-04-09',
                end='2013-04-11',
                as_dataframe=True,
                data_dir=data_dir,
            )
        files_glob = glob.glob(os.path.join(data_dir, '*'))
        assert len(files_glob) == 3



"""
    `National Climatic Data Center`_ `Climate Index Reference Sequential
    (CIRS)`_ drought dataset

    .. _National Climatic Data Center: http://www.ncdc.noaa.gov
    .. _Climate Index Reference Sequential (CIRS): http://www1.ncdc.noaa.gov/pub/data/cirs/
"""

from .core import get_data



import pytest

import ulmo

import test_util


TEST_FILE_PATH = '/tmp/ulmo_test.h5'


def test_get_sites_by_state_code():
    mocked_urls = {
        'http://waterservices.usgs.gov/nwis/dv/':
            'usgs/nwis/RI_daily.xml',
        'http://waterservices.usgs.gov/nwis/iv/':
            'usgs/nwis/RI_instantaneous.xml',
    }

    with test_util.mocked_urls(mocked_urls):
        sites = ulmo.usgs.nwis.get_sites(state_code='RI')
    assert len(sites) > 64


def test_get_sites_by_county_code():
    county_code='51059,51061'
    sites_data_file = 'usgs/nwis/sites_county_%s_daily.xml' % county_code
    with test_util.mocked_urls(sites_data_file):
        sites = ulmo.usgs.nwis.get_sites(county_code=county_code, service='dv')
    assert len(sites) > 23


def test_get_sites_by_huc():
    huc='02070010'
    sites_data_file = 'usgs/nwis/sites_huc_%s_daily.xml' % huc
    with test_util.mocked_urls(sites_data_file):
        sites = ulmo.usgs.nwis.get_sites(huc=huc, service='dv')
    assert len(sites) > 61


def test_get_sites_with_extra_kwarg():
    sites_data_file = 'usgs/nwis/sites_kwarg_agencyCD.xml'
    with test_util.mocked_urls(sites_data_file):
        sites = ulmo.usgs.nwis.get_sites(state_code='TX', agencyCD='USCE', service='dv')
    assert len(sites) == 1


def test_get_sites_single_site():
    site_code = '08068500'
    site_data_file = 'usgs/nwis/site_%s_daily.xml' % site_code
    with test_util.mocked_urls(site_data_file):
        sites = ulmo.usgs.nwis.get_sites(sites=site_code)
    assert len(sites) == 1


def test_get_site_data_single_site():
    site_code = '08068500'
    site_data_file = 'usgs/nwis/site_%s_daily.xml' % site_code
    with test_util.mocked_urls(site_data_file):
        site_data = ulmo.usgs.nwis.get_site_data(site_code, methods='all')
    assert len(site_data) == 26


def test_get_site_data_bad_service_raises_error():
    site_code = '08068500'
    site_data_file = 'usgs/nwis/site_%s_daily.xml' % site_code
    with test_util.mocked_urls(site_data_file):
        with pytest.raises(ValueError):
            ulmo.usgs.nwis.get_site_data(site_code,
                    service="bad_service")


def test_get_site_data_single_site_with_start_and_end():
    site_code = '08068500'
    site_data_file = 'usgs/nwis/site_08068500_instantaneous_2011-11-05_2011-11-18.xml'
    with test_util.mocked_urls(site_data_file):
        site_data = ulmo.usgs.nwis.get_site_data(site_code, start='2011-11-05',
                end='2011-11-18', service='instantaneous',
                methods={'00065': '2'})
    assert len(site_data) == 6
    assert len(site_data['63680:00000']['values']) == 1250


def test_get_site_data_single_site_with_period():
    site_data_file = 'usgs/nwis/site_01117800_instantaneous_P45D.xml'
    site_code = '01117800'
    with test_util.mocked_urls(site_data_file):
        site_data = ulmo.usgs.nwis.get_site_data(site_code, period='P45D',
                service='daily')
    assert len(site_data) >= 1
    assert len(site_data['00060:00003']['values']) == 45


def test_get_sites_multiple_sites():
    site_codes = ['08068500', '08041500']
    sites_data_file = 'usgs/nwis/sites_%s_daily.xml' % '_'.join(site_codes)
    with test_util.mocked_urls(sites_data_file):
        sites = ulmo.usgs.nwis.get_sites(sites=site_codes)
    assert len(sites) == 2


def test_get_sites_by_bounding_box():
    bounding_box_values = '-83.0,36.5,-81.0,38.5'
    sites_data_file = 'usgs/nwis/sites_%s_daily.xml' % bounding_box_values
    with test_util.mocked_urls(sites_data_file):
        sites = ulmo.usgs.nwis.get_sites(bounding_box=bounding_box_values, service='dv')
    assert len(sites) > 244

    
def test_get_sites_by_serving_parameter_code():
    site_code = '08068500'
    parameter_code_value = '00060'
    sites_data_file = 'usgs/nwis/sites_%s_%s_daily.xml' % (site_code, parameter_code_value)
    with test_util.mocked_urls(sites_data_file):
        sites = ulmo.usgs.nwis.get_sites(sites=site_code, parameter_code=parameter_code_value, service='dv')
    assert len(sites) == 1

    
def test_get_sites_by_multiple_serving_parameter_code():
    site_code = '08068500'
    parameter_code_values = '00060,00065'
    sites_data_file = 'usgs/nwis/sites_%s_%s_daily.xml' % (site_code, parameter_code_values)
    with test_util.mocked_urls(sites_data_file):
        sites = ulmo.usgs.nwis.get_sites(sites=site_code, parameter_code=parameter_code_values, service='dv')
    assert len(sites) == 1

def test_get_site_data_multiple_methods():
    site_code = '08054500'
    site_data_file = 'usgs/nwis/site_08054500_multiple_methods.xml'
    with test_util.mocked_urls(site_data_file):
        site_data = ulmo.usgs.nwis.get_site_data(site_code, methods={'62614': 'all', '45592': 'all'})
        assert len(site_data['00054:00003']['values']) == 1
        assert len(site_data.keys()) == 13
